
/*
 * ifs.c
 *
 * get ip address from known interface
 *
 * works on linux/freebsd.
 * Works on Win32 as well!
 */

#ifdef WIN32
#include <winsock2.h>
#include <ws2tcpip.h>
#else
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <unistd.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "urld.h"

/* pry the IP interface address out of the interface
 * outputs:
 * 	the ip address is returned from the interface
 *	0: means we failed utterly for some reason.
 * inputs:
 *	name - name of interface
 */
unsigned long getip(char *name)
{
#ifdef WIN32
    INTERFACE_INFO InterfaceList[10];
    unsigned long nBytesReturned;
    int neededInterface, nNumInterfaces;
#else
	struct ifreq    ifr;
#endif
	int s;
	struct sockaddr_in sin;

	s = socket(AF_INET, SOCK_DGRAM, 0);
	if (s < 0) {
		return(0);
	}

#ifdef WIN32
    /* 
        AFAIK, Windows doesn't associate names like eth0 etc. We are enforcing
        this kind of association, to make the i/f selection more predictable.
        int0 -> Use first ethernet interface
        int1 -> Use second ethernet interface
            and so on. 
    */
    neededInterface = 0; /* Return the first one by default */
    if (name) {
        char *t = strstr(name, "int");
        if (t) 
            t += 3;
        else {
            uDebug(LOG_ERR, "Cannot find '%s'; Please use int0, int1 etc."
                " for interface names on Windows.", name);
            return 0;
        }
        if (t && *t >= '0' && *t <= '9')
            neededInterface = *t - '0';
    } 
    if (WSAIoctl(s, SIO_GET_INTERFACE_LIST, 0, 0, &InterfaceList,
			sizeof(InterfaceList), &nBytesReturned, 0, 0) == SOCKET_ERROR) 
            return 0;

    nNumInterfaces = nBytesReturned / sizeof(INTERFACE_INFO);
    if (neededInterface && neededInterface >= nNumInterfaces) {
        uDebug(LOG_ERR, "Interface %s does not exist.", name);
        return 0;
    }
    if (InterfaceList[neededInterface].iiAddress.Address.sa_family != AF_INET)
        return 0;
    memcpy(&sin, &InterfaceList[neededInterface].iiAddress.Address, sizeof(struct sockaddr));
    if (sin.sin_addr.s_addr == inet_addr("127.0.0.1")) {
        uDebug(LOG_ERR, "Interface '%s' resolves to localhost.", name);
        return 0;
    }
#else
    strncpy(ifr.ifr_name, name, sizeof (ifr.ifr_name));
    if (ioctl(s, SIOCGIFADDR, (caddr_t)&ifr) < 0) {
		close(s);
		return(0);
    }
	close(s);

	/* paranoid check, only want IP
	*/
	if (ifr.ifr_addr.sa_family != AF_INET) 
		return(0);

	bcopy(&ifr.ifr_addr, &sin, sizeof (struct sockaddr));

#endif
	return(sin.sin_addr.s_addr);
}

#ifdef TESTMAIN
main(int argc, char **argv)
{
	unsigned long x;
	char *inet_ntoa();
	struct sockaddr_in sin;

	if (argc != 2) {
		printf("ifs interface\n");
		exit(1);
	}
	x = getip(argv[1]);
	if (x == 0) {
		printf("getip failed, returns 0\n");
		exit(1);
	}
	sin.sin_addr.s_addr = x;
	printf("ip %s\n", inet_ntoa(sin.sin_addr));
}
#endif
