/*
 *  URL Daemon : Project for Winter 2002
 *	Designed by 
 *		Sumit Chawla (sumit@cs.pdx.edu) 
 *	under guidance of
 *		Professor Jim Binkley (jrb@cs.pdx.edu)
 *
 *	File: urlcomm.c
 *
 *	This file contains the communication part of urld. The first version
 *	uses multithreaded implementation, since this seems to be a logical fit to
 *	the problem; sending and receiving are independent of each other and must
 *	be able to proceed simultaneously.
 *	
 *	An easier implementation would have been to simply have two child 
 *	processes, reading or writing based on the configuration file. But for 
 *	interactive mode, we need to be able to do both. And most of the handling
 *	is similar. I am using pthreads, so the code should be easily portable to
 *	other platforms.
 *  
**/

#include <stdio.h>
#ifdef WIN32
#include <windows.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <winsock.h>
#else /* WIN32 */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>
#include <netdb.h>
#include <sys/signal.h>
#ifndef SOCKET
#define SOCKET int
#define closesocket(p) close(p)
#endif /* SOCKET */
#endif /* WIN32 */

#include "urld.h"

#ifdef USE_BPF
#include <fcntl.h>
#include <stdio.h>
#include <sys/uio.h>
#include <sys/time.h>
#include <sys/filio.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <net/bpf.h>
#include <sys/param.h>
#include <net/if.h>
#include <netinet/in_systm.h>
#include <arpa/inet.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <netinet/ip_icmp.h>
#include <netinet/ip_var.h>
#include <netinet/udp_var.h>
#include <netinet/if_ether.h>
#include <net/if_llc.h>
#endif /* USE_BPF */

/* A couple of constants used only in this file */
#define IF_NAME_LEN		200
#define MCAST_ADDR		"224.0.0.1"
#define BCAST_ADDR		"255.255.255.255"

/* Forward declarations */
int tlvToData(char *buffer, int totalLength, PNODE_DATA *data, const char *ipAddr);
int writeToHTML();
int dataToTlv(PNODE_DATA data, char **tlv);
int dumpIfs();
void htmlIntro(FILE *fd);
void htmlClose(FILE *fd);
void close_interfaces();
unsigned long getip(char *name);
#ifdef USE_BPF
static int bpf_open();
#endif

/* 
	The following structure is used only for transmitting data. Even for single-homed
	hosts, we create a dummy URLD_IF that sets isDefault to 1.
 */
typedef struct urld_ifs {
	char if_name[IF_NAME_LEN + 1];
	unsigned long if_ipAddr;
    int none;		/* "null" device declared with no interfaces */
	SOCKET s;
	/* This is a toggle flag; if mcast is 1, bcast is assumed zero. */
	int mcast;		/* send multicast as opposed to broadcast */
#ifdef USE_BPF
	/* Faster to keep all this stuff in the structure */
	int bpf_fd;
#endif /* USE_BPF */
	struct urld_ifs *next;
} URLD_IF, *PURLD_IF;

PURLD_IF ourInterfaces = NULL;

void set_interface(PURLD_IF p);

struct sockaddr_in sinm, sinb;

/* The main TLV, used by writer */
static char *sendTlv = NULL;
static int sendTlvLen = 0;

/* The only socket used for reading. We read from both mcast and bcast */
static int socketRead = 0;


void send_broadcast(PURLD_IF p)
{
	extern int errno;

#ifdef USE_BPF
	int usebpf = 1;
#else
	int usebpf = 0;
#endif
		
	/* Just send to 255.255.255.255 on whatever interface.
	*/
	if (p->none || !usebpf) {
		char *sender;
		uDebug(LOG_INFO, "ordinary sendto bcast write\n");
		if (sendto(p->s, sendTlv, sendTlvLen, 0, 
			(struct sockaddr *)&(sinb), sizeof(sinb)) < 0) {
			sender = inet_ntoa(sinb.sin_addr);
			uDebug(LOG_WARN, "error: (%s): Could not send broadcast %s to %s", strerror(errno), sender, p->if_name);
		}
		return;
	}
#ifdef USE_BPF
	/* Use BPF for broadcast 
	*/
	uDebug(LOG_INFO, "bpf bcast write\n");
	if (writebpf(sendTlv, sendTlvLen, p) < 0) {
		uDebug(LOG_WARN, "Could not broadcast to %s", p->if_name);
	}
#endif /* USE_BPF */
}

int writeOnce() 
{
	PURLD_IF p = ourInterfaces;

	/* loop thru interfaces
	 */
	while (p) {

		/* do any dynamic setup needed for interface
		*/
		set_interface(p);

		/* if broadcast
		*/
		if (!p->mcast) {
			send_broadcast(p);
			p = p->next;
			continue;
		}
		/* multicast case
		*/
		uDebug(LOG_INFO, "ordinary sendto mcast write\n");
		if (sendto(p->s, sendTlv, sendTlvLen, 0, 
			(struct sockaddr *)&(sinm), sizeof(sinm)) < 0) {
			uDebug(LOG_WARN, "Could not send multicast to %s", p->if_name);
		}
		p = p->next;
	}
	return 0;
}

int dataAvailable()
{
	int rc;
	fd_set set;

	FD_ZERO(&set);
	FD_SET(socketRead, &set);
	rc = select(socketRead + 1, &set, NULL, NULL, NULL);

	return rc;
}

static char inMsg[URL_MAXMSG]; /* Anything more than MAXMSG will be truncated, so beware! */

int readOnce()
{
	/* Read from socket and insert in the node information */
	int fromlen;
	struct sockaddr_in from;
	int rc;
	char *sender;
	extern int errno;

	fromlen = sizeof(from);
	rc = recvfrom(socketRead, inMsg, URL_MAXMSG, 0, (struct sockaddr *)&from, &fromlen);
	if (rc < 0)
	{
		/* Unknown failure */
#ifdef WIN32
		uDebug(LOG_ERR, "Error from recvfrom = %d", WSAGetLastError());
#else
		uDebug(LOG_ERR, "recvfrom failed due to %s.", strerror(errno));
#endif /* WIN32 */
		return 0;
	}

	/* First step: we search for the guy who sent this */
	sender = inet_ntoa(from.sin_addr);
	if (0 != tlvToData(inMsg, rc, &remoteData, sender))
	{
		uDebug(LOG_WARN, "An incoming message was discarded.");
		/* The illegal messages are thrown away. Log this and go on. */
	}
	return 0;
}

#ifdef THREADED_MODEL

/* Using GNU/Win32 Pthread implementation for Windows */

/* The mutex that decides who gets the control of URL list */
pthread_t writerThreadId, readerThreadId, walkerThreadId;
pthread_attr_t attr;
pthread_mutex_t listMutex = PTHREAD_MUTEX_INITIALIZER;


void make_cancellable()
{
#ifdef PTHREAD_CANCEL_ASYNCHRONOUS
	int cancel_type;
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, &cancel_type);
#endif /* PTHREAD_CANCEL_ASYNCHRONOUS */
}

void *writerThread(void *param)
{
	/* Let this thread be cancelled asynchronously. */
	make_cancellable();
	/* Very simple. We can use some better mechanism for termination in the next release. */
	while (1)
	{
		uDebug(LOG_INFO, "Sending URLs..");
		writeOnce();
		sleep(sendTime);
	}
	uDebug(LOG_INFO, "Terminating the writer thread");
	return NULL;
}

void *readerThread(void *param)
{
	/* Let this thread be cancelled asynchronously. */
	make_cancellable();
	/* Very simple. We can use some better mechanism for termination in the next release. */
	while (1)
	{
		/* Block only if we have data available. */
		if (dataAvailable())
		{
			pthread_mutex_lock(&listMutex);
			readOnce();
			pthread_mutex_unlock(&listMutex);
		}
	}
	uDebug(LOG_INFO, "Terminating the reader thread");
	return NULL;
}

void *walkerThread(void *param)
{
	/* Let this thread be cancelled asynchronously. */
	make_cancellable();
	/* Very simple. We can use some better mechanism for termination in the next release. */
	while (1)
	{
		sleep(expireTime);
		pthread_mutex_lock(&listMutex);
		writeToHTML();
		pthread_mutex_unlock(&listMutex);
	}
	uDebug(LOG_INFO, "Terminating the walker thread");
	return NULL;
}
#endif /* THREADED_MODEL */


/*
 *	There are two activities that the reader needs to do. The first is to 
 *	listen for incoming UDP messages and record what the changes are. The
 *	second is a periodic update of the HTML file. These activities should
 *	be able to proceed simultaneously. Hence I am using two dedicated 
 *	threads to do these two jobs.
**/
int startupReader()
{
	int fromlen;
	struct sockaddr_in server;
	int one = 1;
	struct ip_mreq mreq;


	if (urlMode & URL_MODE_READ) 
	{
		/* Let's skip using BPF for reading, unless we need to. */
		server.sin_family = AF_INET;
		server.sin_addr.s_addr = INADDR_ANY;   /* OK from any machine */
		server.sin_port = htons(recvPort);      /* specific port */

		fromlen = sizeof( struct sockaddr_in );

		/* Allow others to bind to this port if needed */
		setsockopt(socketRead, SOL_SOCKET, SO_REUSEADDR, (char *)&one, sizeof(one));

		/* bind protocol to socket
		*/
		if (bind(socketRead, (struct sockaddr *)&server, sizeof(server))) {
			uDebug(LOG_ERR, "Could not bind socket for reading");
			return -1;
		}
		/* Allow multicast requests to be received on the specified 
		 * interface 
		*/
		mreq.imr_multiaddr.s_addr = inet_addr(MCAST_ADDR);
		mreq.imr_interface.s_addr = INADDR_ANY;
		if (setsockopt(socketRead, IPPROTO_IP, IP_ADD_MEMBERSHIP,
			(char*)&mreq, sizeof(mreq)) < 0) {
			uDebug(LOG_ERR, "Could not bind to the multicast address");
		}

#ifdef THREADED_MODEL
		/* Start the reader threads */
        pthread_attr_init(&attr);
        pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);

		/* First, the socket handler who reads incoming messages */
		if (pthread_create(&readerThreadId, &attr, readerThread, NULL)) {
			uDebug(LOG_ERR, "Could not create the reader thread!");
			return -1;
		}

		if (pthread_create(&walkerThreadId, &attr, walkerThread, NULL)) {
			uDebug(LOG_ERR, "Could not create the walker thread!");
			return -1;
		}
		return 0;
#else /* THREADED_MODEL */
		/* The rest of the stuff is driven through the timer */
		return 0;
#endif /* THREADED_MODEL */
	}
	return 0;
}

/*
 *	The writer is pretty straightforward in this initial release; all it needs
 *	to do is keep writing the stuff it read from the configuration file. It 
 *	doesn't need to check for changes to the configuration file at the moment.
 *	But in order for the writer to proceed along with reader, the easiest 
 *	design is to use an independent thread.
**/
int startupWriter()
{
	int one = 1;
	FILE *fd;

	if (urlMode & URL_MODE_WRITE)
	{
		/* First, sanity check */
		if (myData == NULL || myData->urls == NULL) {
			uDebug(LOG_ERR, "You must specify at least one URL to write to!");
			return -1;
		}

		myData->nodeType = senderType;

		strcpy(myData->name, senderName);

		/* Convert the data to the format we need for transmission */
		sendTlvLen = dataToTlv(myData, &sendTlv);
		if (sendTlvLen <= 0)
		{
			uDebug(LOG_ERR, "Aborting the writer startup");
			return -1;
		}
		else if (sendTlvLen > URL_MAXMSG)
		{
			uDebug(LOG_ERR, "Msg length (%d) exceeds max allowed length is %d", sendTlvLen, URL_MAXMSG);
			return -1;
		}



		/* multicast socket setup with port and mcast ip
		*/ 
		sinm.sin_port = htons(sendPort);
		sinm.sin_family = AF_INET;
		sinm.sin_addr.s_addr = inet_addr(MCAST_ADDR);

		/* broadcast socket setup with port and mcast ip
		*/ 
		/* note: BPF doesn't use sinb, it uses writebpf instead 
		 * But if no interface is specified, we have to use sinb 
		*/
		sinb.sin_port = htons(sendPort);
		sinb.sin_family = AF_INET;
		sinb.sin_addr.s_addr = inet_addr(BCAST_ADDR);

		/* Write the initial file stuff */
		fd = fopen(recvPath, "w");
		if (fd == NULL)
		{
			uDebug(LOG_ERR, "Cannot open %s for writing", recvPath);
			return -1;
		}
		htmlIntro(fd);
		htmlClose(fd);
		fclose(fd);

		/* Start the writer */
#ifdef THREADED_MODEL
        pthread_attr_init(&attr);
        pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);
		if (pthread_create(&writerThreadId, &attr, writerThread, NULL)) {
			uDebug(LOG_ERR, "Could not create the writer thread!");
			return -1;
		}
		return 0;
#else /* THREADED_MODEL */
		/* The rest of the stuff is driven through the timer */
		return 0;
#endif /* THREADED_MODEL */
	}
	return 0;
}

/*
 *	We stop the two reader threads here, and kill the HTML file. 
*/
int stopReader()
{
	if (urlMode & URL_MODE_READ) 
	{
#ifdef THREADED_MODEL
#ifdef PTHREAD_CANCEL_ASYNCHRONOUS
		pthread_cancel(readerThreadId);
#endif /* PTHREAD_CANCEL_ASYNCHRONOUS */
#endif /* THREADED_MODEL */
		/* Release all the nodes we have */
		if (remoteData) {
			PNODE_DATA p = remoteData;
			while (p) {
				PURL_DATA u = p->urls;

				while (u) {
					p->urls = u->next;
					free(u);
					u = p->urls;
				}

				remoteData = p->next;
				free(p);
				p = remoteData;
			}
		}
		/* Close the read socket as well */
		closesocket(socketRead);
	}
	return 0;
}

/*
 *	We stop the writer thread here.
*/
int stopWriter()
{
	if (urlMode & URL_MODE_WRITE)
	{
#ifdef THREADED_MODEL
#ifdef PTHREAD_CANCEL_ASYNCHRONOUS
		pthread_cancel(writerThreadId);
#endif /* PTHREAD_CANCEL_ASYNCHRONOUS */
#endif /* THREADED_MODEL */
		/* Release the data stored by the Writer */
		close_interfaces();
		/* Release all the nodes we have */
		if (myData) {
			PNODE_DATA p = myData;
			while (p) {
				PURL_DATA u = p->urls;

				while (u) {
					p->urls = u->next;
					free(u);
					u = p->urls;
				}

				myData = p->next;
				free(p);
				p = myData;
			}
		}
		/* The rest of the stuff is driven through the timer */
		return 0;
	}
	return 0;
}

/*
 * Add the interface to the list of interfaces we know (and care) about.
 */
int addIf(char *ifName, char *mode)
{
	PURLD_IF puif;
	puif = (PURLD_IF) malloc(sizeof(URLD_IF));
    if (puif == NULL) {
       uDebug(LOG_ERR, "malloc in addIf failed");
       exit (-1);
	}
	strncpy(puif->if_name, ifName, IF_NAME_LEN);
	puif->if_name[IF_NAME_LEN] = 0;
	puif->none = 0;
    if (strcmp(ifName, "none") == 0) {
		puif->none = 1;
	}
	puif->if_ipAddr = 0;	/* must be set at runtime */
	puif->next = NULL;
	puif->mcast = strchr(mode, 'm') ? 1:0;
	if (ourInterfaces)
		puif->next = ourInterfaces;
	ourInterfaces = puif;
	return 0;
}

int dumpIfs()
{
	PURLD_IF p = ourInterfaces;

	uDebug(LOG_INFO, "Using the following interfaces for sendto:");
	while (p) {
		char *addr = inet_ntoa(*(struct in_addr *)&p->if_ipAddr);
		uDebug(LOG_INFO, "%s %s %s", p->if_name, addr,
			p->mcast ? "MULTICAST": "BROADCAST");
		p = p->next;
	}
	return 0;
}

/*
 * allocate sockets and setup interfaces in general 
 * one time only.
 *
 * There is a write fd per interface.
 * There is only one global fd for reading.
 *
 * One time setsockopts are done here too.
 */
int init_interfaces()
{
	PURLD_IF header = ourInterfaces;
	int one = 1;
	int mttl = 1;

	if (header == NULL) 
		return -1;

	while (header) {
		header->s = socket(AF_INET, SOCK_DGRAM, 0);
		if (header->s < 0) {
			uDebug(LOG_ERR, "Couldn't create socket for %s", header->if_name);
			return -1;
		}
#ifdef USE_BPF
		if (!header->none)
			if (openbpf(header) < 0) 
				return -1;
#endif /* USE_BPF */
		/* setup onetime options as needed, for broadcast
		 * or multicast
		 */ 
		if (!header->mcast) {
			if (setsockopt(header->s, SOL_SOCKET, 
				SO_BROADCAST, (char *)&one, sizeof(one)) < 0) {
			   uDebug(LOG_ERR, "SO_BROADCAST failed for interface %s.", header->if_name);
			}
#ifndef USE_BPF
#ifdef SO_BINDTODEVICE
			/* Use the SO_BINDTODEVICE on linux to bind
			 * broadcast or whatever to interface
			 */
			if (setsockopt(header->s, SOL_SOCKET, SO_BINDTODEVICE,
				&(header->if_name), sizeof(header->if_name)) < 0) {
				uDebug(LOG_ERR, "Couldn't set SO_BINDTODEVICE for %s!", 
                    header->if_name);
			} 
#else /* SO_BINDTODEVICE */
            {
                struct sockaddr_in ourIf;
                ourIf.sin_family = AF_INET;
                ourIf.sin_port = htons(sendPort);
                ourIf.sin_addr.s_addr = header->if_ipAddr;
                /* Don't have SO_BINDTODEVICE.. Try bind instead */
                if (bind(header->s, (struct sockaddr *)&ourIf, 
                    sizeof(struct sockaddr))) {
				    uDebug(LOG_ERR, "Couldn't bind to %s!", header->if_name);
                }
            }
#endif /* SO_BINDTODEVICE */
#endif /* USE_BPF */
		}
		header = header->next;
	}
	if (urlMode && URL_MODE_READ) {
		socketRead = socket(AF_INET, SOCK_DGRAM, 0);
		if (socketRead < 0) {
			uDebug(LOG_ERR, "Couldn't create socket for reading");
			return -1;
		}
	}
	return 0;
}

/*
 * this routine is for writing.  aim socket
 * at a given interface.
 */
void set_interface(PURLD_IF p)
{
	struct sockaddr_in sin;
	int one=1;

	/* bail if we don't have any interface
	*/
	if (p->none) {
		return;
	}

	/* get/set interface ip address from interface
	*/
	p->if_ipAddr = sin.sin_addr.s_addr = getip(p->if_name);
    if (p->if_ipAddr == 0) {
		uDebug(LOG_ERR, "could not get a valid IP address for interface %s", 
			p->if_name);
		exit(1);
	}

	if (debug) {
		printf("set_interface: if name %s, ip %s\n", 
			p->if_name, inet_ntoa(sin.sin_addr));
	}

	if (p->mcast) {
		if (setsockopt(p->s, IPPROTO_IP, IP_MULTICAST_IF, 
			(char *)&(p->if_ipAddr), sizeof(p->if_ipAddr)) < 0) {
			uDebug(LOG_ERR, "Couldn't set IP_MULTICAST_IF for %s!", p->if_name);
		}
        if (setsockopt(p->s, IPPROTO_IP, IP_MULTICAST_TTL, 
			(char *)&one, sizeof(one)) < 0) {
        	uDebug(LOG_ERR, "IP_MULTICAST_TTL failed for interface %s.", p->if_name);
		}
	}
	else {
#ifdef USE_BPF
		/* already taken care of due to per interface socket descriptor
		*/
		return;
#else
#ifdef linuxBUTNOTUSED
		/* use SO_BINDTODEVICE passing in the interface string name
		*/
		if( setsockopt(p->s, SOL_SOCKET, SO_BINDTODEVICE, 
				p->if_name, IF_NAME_LEN)) {
        		uDebug(LOG_ERR, "SO_BINDTODEVICE failed for interface %s.", 
					p->if_name);
		}
	    if (debug) {
		    printf("SO_BINDTODEVICE called on %s\n", p->if_name);
	    }
#endif
#endif
	}
}

void close_interfaces()
{
	PURLD_IF header = ourInterfaces;

	while (header)
	{
		closesocket(header->s);
#ifdef USE_BPF
		if (!header->none)
			closebpf(header);
#endif /* USE_BPF */
		ourInterfaces = header->next;
		free(header);
		header = ourInterfaces;
	}
	ourInterfaces = NULL;
}

#ifdef USE_BPF

static int bpf_bufsize;
#define BPFBUFSIZE 32768
static char bpf_buf[BPFBUFSIZE];

/* type of interface */
static int if_op;

/* Link-level header length */
#ifndef ETHER_HDRLEN
#define ETHER_HDRLEN    sizeof(struct ether_header)
#endif /* ETHER_HDRLEN */


/* 
	Since the port number is configurable, we need to correct this
	filter AFTER the correct port value has been loaded.
	Filtering is as follows:
	1) Check if it is a UDP packet
	2) Check if destination port matches or not

*/
/* ethhdr - 14 octets, iphdr - 20 octets */
static struct bpf_insn udp_filter[] = {
		/* A <- P[23:1] - load byte - IP protocol */
		BPF_STMT(BPF_LD+BPF_B+BPF_ABS, 14 + 9),
		/* pc += (A == IPPROTO_ICMP) ? 0 : 3 */
		BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, IPPROTO_UDP, 0, 3),
		/* A <- P[34:1] - load word - Port number */
		BPF_STMT(BPF_LD+BPF_H+BPF_ABS, 14 + 20),
		/* pc += (A == recvPort) ? 0 : 1 (agentadv) */
		BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, 43333, 0, 1),
		BPF_STMT(BPF_RET+BPF_K, (u_int)-1), /* accept the packet */
		BPF_STMT(BPF_RET+BPF_K, 0) /* filter packet */
};

static int udp_filter_len = sizeof udp_filter / sizeof (struct bpf_insn);


/* Stuff for writing to 255.255.255.255 using BPF */
int openbpf(PURLD_IF pif)
{
	int rc;
	int fd;
	int v = BPFBUFSIZE;
 	struct ifreq ifr;
	struct bpf_program p;

	/* Correct the port number manually. */
	udp_filter[3].k = htons(recvPort);

	p.bf_len = udp_filter_len;
	p.bf_insns = udp_filter;

	fd = bpf_open();
	if (fd < 0)
		return(-1);

	/* set the buffer size, must match read size
	*/
    (void) ioctl(fd, BIOCSBLEN, (caddr_t)&v);

	/* set the network device
	*/
    (void)strncpy(ifr.ifr_name, pif->if_name, sizeof(ifr.ifr_name));
    if (ioctl(fd, BIOCSETIF, (caddr_t)&ifr) < 0) 
	{
        uDebug(LOG_ERR, "%s: %s", pif->if_name, strerror(errno));
		close(fd);
		return(-1);
    }

	/* The write area is kept open even if we never write. */
	if (urlMode & URL_MODE_READ) {

		/* Get the data link layer type. */
		if (ioctl(fd, BIOCGDLT, (caddr_t)&v) < 0) {
			uDebug(LOG_ERR, "BIOCGDLT: %s", strerror(errno));
			close(fd);
			return(-1);
		}
		if_op = v;

		/* turn on immediate mode so we don't hang in reads
		 * and get 1 packet at a time. overkill on that there buffer,
		 * but oh well (size is too big).
		*/
		v = 1;
		rc = ioctl(fd, BIOCIMMEDIATE, &v);
		if (rc < 0) {
			uDebug(LOG_ERR,  "biocimmediate: %s", strerror(errno));
			close(fd);
			return(-1);
		}
		rc = ioctl(fd, BIOCGBLEN, &bpf_bufsize);
		if (rc < 0) {
			uDebug(LOG_ERR,  "biocgblen: %s", strerror(errno));
			close(fd);
			return(-1);
		}
		/* install the local packet filter for reads
		*/
		rc = ioctl(fd, BIOCSETF, &p);
		if (rc < 0) {
			uDebug(LOG_ERR, "biocsetf: %s", strerror(errno));
			close(fd);
			return(-1);
		}
	} else {
		shutdown(fd, 0);	/* shut down read */
	}
	pif->bpf_fd = fd;
	return 0;
}

int closebpf(PURLD_IF pif)
{
	close(pif->bpf_fd);
	return 0;
}

int writebpf(char *packet, int length, PURLD_IF pif)
{
	int bufoffset=0;
	unsigned char buf[1514];
	struct iovec iov[2];
	int rc; 
	extern int debug;
	struct ether_header *eh;
	struct udpiphdr *ui;
	char macsrc[6];
	struct in_addr ipsrc, ipdst;

	/* copy payload into same packet buffer, skipping layers 2/3/4
	*/
	char *pkt = (char *) &buf[0];
	pkt = pkt + sizeof(struct ether_header) + sizeof(struct udpiphdr);	
	bcopy(packet, pkt, length);	/* copy data part into pkt part */

	/* build the ethernet header first.
	*/
	eh = (struct ether_header *)&buf[0];
	/* We spoof the MAC address */
	memset(eh->ether_shost, 0xff, 6);
	memset(eh->ether_dhost, 0xff, 6);
	/* The destination is for broadcast, source is current interface address */
	ipdst.s_addr = 0xffffffff;
	/* TBD. fixme.  get ip address dynamically
	*/
	ipsrc.s_addr = pif->if_ipAddr;

	eh->ether_type = htons(ETHERTYPE_IP);
	bufoffset = ETHER_HDRLEN;	/* length so far */

    /*
     * Fill in mbuf with extended UDP header
     * and addresses and length put into network format.
     */
    ui = (struct udpiphdr *) &buf[bufoffset];
    bzero(ui->ui_x1, sizeof(ui->ui_x1));	/* zero out-non csum 1st 9 ip bytes */
    ui->ui_pr = IPPROTO_UDP;
    ui->ui_len = htons((u_short)length + sizeof (struct udphdr));
    ui->ui_src = ipsrc;	/* already byte-swapped */
    ui->ui_dst = ipdst;	/* already byte-swapped */

    ui->ui_sport = htons(sendPort);
    ui->ui_dport = htons(sendPort);
    ui->ui_ulen = ui->ui_len;

    /*
     * do checksum
     */
    ui->ui_sum = 0;
    if ((ui->ui_sum = in_cksum(ui, sizeof (struct udpiphdr) + length)) == 0)
            ui->ui_sum = 0xffff;

	/* udp header is finished, ip header needs work
	*/

        ((struct ip *)ui)->ip_v = 4;
        ((struct ip *)ui)->ip_hl = 5;
        ((struct ip *)ui)->ip_tos = IPTOS_LOWDELAY;
        ((struct ip *)ui)->ip_len = htons(sizeof (struct udpiphdr) + length);
        ((struct ip *)ui)->ip_id = 0;
        ((struct ip *)ui)->ip_off = 0;
        ((struct ip *)ui)->ip_ttl = 16;
	/* one more cksum calculation on ip hdr this time
	*/
        ((struct ip *)ui)->ip_sum = 0;
        ((struct ip *)ui)->ip_sum = in_cksum ((struct ip *)ui, sizeof(struct ip));

	bufoffset += sizeof(struct udpiphdr);
	bufoffset += length;	/* now we have total length */
	uDebug(LOG_INFO, "bpf write, total length %d\n", bufoffset);

	iov[0].iov_base = buf;
	iov[0].iov_len = bufoffset;
	rc = writev(pif->bpf_fd, iov, 1);
	uDebug(LOG_INFO, "bpf write returns %d\n", rc);
	return(rc);
}

/*
 * in_cksum --
 *	Checksum routine for Internet Protocol family headers (C Version)
 *	(from BSD ping code)
 */
int in_cksum(u_short *addr, int len)
{
	register int nleft = len;
	register u_short *w = addr;
	register int sum = 0;
	u_short answer = 0;

	/*
	 * Our algorithm is simple, using a 32 bit accumulator (sum), we add
	 * sequential 16 bit words to it, and at the end, fold back all the
	 * carry bits from the top 16 bits into the lower 16 bits.
	 */
	while (nleft > 1)  {
		sum += *w++;
		nleft -= 2;
	}

	/* mop up an odd byte, if necessary */
	if (nleft == 1) {
		*(u_char *)(&answer) = *(u_char *)w ;
		sum += answer;
	}

	/* add back carry outs from top 16 bits to low 16 bits */
	sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
	sum += (sum >> 16);			/* add carry */
	answer = ~sum;				/* truncate to 16 bits */
	return(answer);
}

/*
 * bpf_open
 *	find a bpf device
 *	open rd/wr
 *
 * note: opened read/write !!!
 *
 * returns:
 *	valid fd or -1, plus str errno in errbuf
 */ 

static int bpf_open()
{
	int fd;
	int n = 0;
	extern int errno;
	char device[sizeof "/dev/bpf000"];

	/*
	 * Go through all the minors and find one that isn't in use.
	 */
	do {
			(void)sprintf(device, "/dev/bpf%d", n++);
			fd = open(device, O_RDWR);
	} while (fd < 0 && errno == EBUSY);

	/*
	 * XXX better message for all minors used
	 */
	if (fd < 0)
			uDebug(LOG_ERR, "%s: %s", device, strerror(errno));

	return (fd);
}

#endif /* USE_BPF */
