/*
 *	URL Daemon : Project for Winter 2002
 *	Designed by 
 *		Sumit Chawla (sumit@cs.pdx.edu) 
 *	under guidance of
 *		Professor Jim Binkley (jrb@cs.pdx.edu)
 *
 *	File: urlconfig.c
 *
 *	This file contains configuration functions for urld.
 *
 *	The routines in this file control the interfacing of urld with the 
 *	configuration file. Reading from the configuration file, we initialize
 *	all the internal data structures. The communication is not done from 
 *	any of the functions in this file, though.
 *
 * note: ts=4
**/
#include <stdio.h>
#include <string.h>
#ifdef WIN32
#include <direct.h>
#define strcasecmp stricmp
#endif
#include "urld.h"

/* Forward declarations */
int addURL(const char *url, const char *desc);
int dumpURLs(PNODE_DATA nodeHead);
int addIf(char *ifName, char *mode);
int dumpIfs();


int configRead(char *configName)
{
	FILE			*fd;
	char            linebuf[1024];
	char            opbuf[1024];
	char			p1[1024];
	char			p2[1024];
	char			p3[1024];
	char			*opword;
	char			*result;
	int             lineNumber = 1;
	char			*restOfLine;
	int				interfaces = 0;		/* interface count */

	recvPath[0] = 0;
	senderType = NODE_TYPE_UNK;

   /*
	* setup a default or two
	*/
#ifndef WIN32
    strcpy(recvPath, "/tmp/urld.html");
#else
    /* Look for the following in order: TEMP, TMP, current dir. */
    restOfLine = getenv("TEMP");
    if (!restOfLine) restOfLine = getenv("TMP");
    if (!restOfLine) restOfLine = getcwd(p2, sizeof(p2));
    if (!restOfLine) restOfLine = "c:\\temp";
    sprintf(recvPath, "%s\\urld.html", restOfLine);
#endif

	fd = fopen(configName, "r");
	if (fd == NULL) {
		uDebug(LOG_ERR,
		       "ConfigRead(): fopen of config file %s failed",
		       configName);
		return (-1);
	}

	/*
	 * while not at EOF read a line, parse it
	 */
	while (feof(fd) == 0) {
		result = fgets(linebuf, 1023, fd);
		if (result == NULL)
			break;
		/*
		 * skip comment
		 */
		if (linebuf[0] == '#') {
			lineNumber++;
			continue;
		}
		/*
		 * skip empty lines
		 */
		if (linebuf[0] == '\n' || linebuf[0] == '\r') {
			lineNumber++;
			continue;
		}
		/*
		 * skip whitespace in line
		 */
		while (*result == ' ' || *result == '\t') {
			result++;
		}
		/*
		 * find first token
		 */
		strncpy(opbuf, result, 1024);
		opword = strtok(opbuf, " \t");
		if (opword == NULL || (strlen(opword) == 0))
			continue;
		/*
			We do the switch from within this function, instead of calling
			an external switchOnOp. 
		*/
		if (strcasecmp(opword, "mode") == 0) {
			/* mode <mode> */
			sscanf(result, "%s %s", p1, p2);
			if (strlen(p2) == 0) {
				uDebug(LOG_ERR, "Line %d of %s: mode must be defined", lineNumber, configName);
				return -1;
			}
			urlMode = atoi(p2);
			if (urlMode <= 0) {
				uDebug(LOG_ERR, "Line %d of %s: mode must be a positive value", lineNumber, configName);
				return -1;
			}
		    if (urlMode > 3) {
 				urlMode = 3;
	        }
		} else if (strcasecmp(opword, "syslabel")==0) {
			/* syslabel <Node type> <Node Name> */
			sscanf(result, "%s %s %s", p1, p2, p3);
			if (strlen(p2) == 0 || strlen(p3) == 0) {
				uDebug(LOG_ERR, "Line %d of %s: sendIP must have a node type & node name", lineNumber, configName);
				return -1;
			}
			if (strcasecmp(p2, "fixed") == 0) {
				senderType = NODE_TYPE_HA;
			} 
			else if (strcasecmp(p2, "mobile") == 0) {
				senderType = NODE_TYPE_MNODE;
			} 
			else {
				senderType = atoi(p2);
			}
			if (senderType <= 0) {
				uDebug(LOG_ERR, "Line %d of %s: sender type should be fixed or mobile", lineNumber, configName);
				return -1;
			}
			if (senderType > 2) {
				uDebug(LOG_ERR, "Line %d of %s: sender type should be fixed or mobile", lineNumber, configName);
				return -1;
			}
			restOfLine = strstr(result, p2);
			restOfLine += (strlen(p2) + 1);
			/* The next line must NEVER happen!!! */
			if (restOfLine == NULL) restOfLine = p3;
			strncpy(p3, restOfLine, 1024);
			/* Truncate the name at NODE_NAME_MAX length */
			p3[NODE_NAME_MAX] = 0;
			/* Remove the trailing newline */
			restOfLine = strchr(p3, '\n');
			if (restOfLine) *restOfLine = 0;
			strcpy(senderName, p3);
		} else if (strcasecmp(opword, "port") == 0) {
			/* port <UDP port number> */
			sscanf(result, "%s %s", p1, p2);
			if (strlen(p2) == 0) {
				uDebug(LOG_ERR, "Line %d of %s: sendPort must be defined", lineNumber, configName);
				return -1;
			}
			recvPort = sendPort = atoi(p2);
			if (sendPort <= 0) {
				uDebug(LOG_ERR, "Line %d of %s: port must be a positive value", lineNumber, configName);
				return -1;
			}
		} else if (strcasecmp(opword, "send") == 0) {
			/* send <URL> <Description> */
			sscanf(result, "%s %s %s", p1, p2, p3);
			if (strlen(p2) == 0 || strlen(p3) == 0) {
				uDebug(LOG_ERR, "Line %d of %s: 'send' must specify both URL & Description", lineNumber, configName);
				return -1;
			}
			restOfLine = strstr(result, p2) + strlen(p2);
           	restOfLine = strstr(restOfLine, p3);

			/* The next line must NEVER happen!!! */
			if (restOfLine == NULL) restOfLine = p3;
			strncpy(p3, restOfLine, URL_DESC_MAX);
			/* Limit the sizes of the two strings */
			p3[URL_DESC_MAX] = 0;
			p2[URL_SIZE_MAX] = 0;
			/* Remove the trailing newline */
			restOfLine = strchr(p3, '\n');
			if (restOfLine) *restOfLine = 0;
			/* Now we have the html and description in p2 and p3. Create a new node */
			addURL(p2, p3);
		} else if (strcasecmp(opword, "interface") == 0) {
			/* interface <interface-name> [mode] */
			sscanf(result, "%s %s %s", p1, p2, p3);
			if (strlen(p2) == 0 ) {
				uDebug(LOG_ERR, "Line %d of %s: 'interface' must specify interface name", lineNumber, configName);
				return -1;
			}
			/* do broadcast (write) by default.
	         */
			if (strlen(p3) == 0) {
				restOfLine = strchr(p3, '\n');
				strcpy(p3, "b");
			} else
				restOfLine = strchr(p3, '\n');
			if (restOfLine) 
				*restOfLine = 0;
			/* Are we trying for BOTH m and b? 
			*/
			if (strchr(p3, 'm') && strchr(p3, 'b')) {
				uDebug(LOG_ERR, "Line %d of %s: Doing both multicast and broadcast is not allowed", lineNumber, configName);
				return -1;
			}
			/* Add this to the list of interfaces we use. */
			addIf(p2, p3);
			interfaces++;
		} else if (strcasecmp(opword, "sendTime") == 0) {
			/* sendTime <seconds> */
			sscanf(result, "%s %s", p1, p2);
			if (strlen(p2) == 0) {
				uDebug(LOG_ERR, "Line %d of %s: sendTime must be defined", lineNumber, configName);
				return -1;
			}
			sendTime = atoi(p2);
			if (sendTime <= 0) {
				uDebug(LOG_ERR, "Line %d of %s: sendTime cannot be <= zero", lineNumber, configName);
				return -1;
			}
		} else if (strcasecmp(opword, "recv") == 0) {
			/* recv <complete path & name of html file> */
			sscanf(result, "%s %s", p1, p2);
			if (strlen(p2) == 0) {
				uDebug(LOG_ERR, "Line %d of %s: recv must not be blank", lineNumber, configName);
				return -1;
			}
			restOfLine = strstr(result, p2);
			/* The next line must NEVER happen!!! */
			if (restOfLine == NULL) restOfLine = p2;
			strncpy(recvPath, restOfLine, 1023);
			recvPath[1023] = 0;
			/* Remove the trailing newline */
			restOfLine = strchr(recvPath, '\n');
			if (restOfLine) *restOfLine = 0;
		} else if (strcasecmp(opword, "expire") == 0) {
			/* expire <seconds> */
			sscanf(result, "%s %s", p1, p2);
			if (strlen(p2) == 0) {
				uDebug(LOG_ERR, "Line %d of %s: expire must be defined", lineNumber, configName);
				return -1;
			}
			expireTime = atoi(p2);
			if (expireTime <= 0) {
				uDebug(LOG_ERR, "Line %d of %s: expire cannot be <= zero", lineNumber, configName);
				return -1;
			}
		} else if (strcasecmp(opword, "urlTime") == 0) {
			/* urlTime <seconds> */
			sscanf(result, "%s %s", p1, p2);
			if (strlen(p2) == 0) {
				uDebug(LOG_ERR, "Line %d of %s: urlTime must be defined", lineNumber, configName);
				return -1;
			}
			urlTime = atoi(p2);
			if (urlTime <= 0) {
				uDebug(LOG_ERR, "Line %d of %s: urlTime cannot be <= zero", lineNumber, configName);
				return -1;
			}
		} else if (strcasecmp(opword, "MaxURLs") == 0) {
			/* MaxURLs <Maximum number of URLs per system> */
			sscanf(result, "%s %s", p1, p2);
			if (strlen(p2) == 0) {
				uDebug(LOG_ERR, "Line %d of %s: MaxURLs must be defined", lineNumber, configName);
				return -1;
			}
			maxURLs = atoi(p2);
			if (maxURLs <= 0) {
				uDebug(LOG_ERR, "Line %d of %s: MaxURLs cannot be <= zero", lineNumber, configName);
				return -1;
			}
		} else if (strcasecmp(opword, "MaxSystems") == 0) {
			/* MaxSystems <Maximum number of Systems per page> */
			sscanf(result, "%s %s", p1, p2);
			if (strlen(p2) == 0) {
				uDebug(LOG_ERR, "Line %d of %s: MaxSystems must be defined", lineNumber, configName);
				return -1;
			}
			maxSystems = atoi(p2);
			if (maxSystems <= 0) {
				uDebug(LOG_ERR, "Line %d of %s: MaxSystems cannot be <= zero", lineNumber, configName);
				return -1;
			}
		} else {
			uDebug(LOG_ERR, "Unknown line %d in configuration file %s", lineNumber, configName);
			return -1;
		}
		lineNumber++;
	}
	fclose(fd);

	/* Now do sanity check */
	if (urlMode == URL_MODE_DISABLE) {
		uDebug(LOG_ERR, 
			"Could not locate a valid mode in configuration file %s",
										configName);
		return -1;
	}
	
	if (urlMode & URL_MODE_READ) {
		if (strlen(recvPath) == 0) {
			uDebug(LOG_ERR, "The output HTML path must be specified!");
			return -1;
		}
		/* We must have a valid file we can create */
		fd = fopen(recvPath, "w");
		if (fd == NULL) {
			uDebug(LOG_ERR, 
				"Could not open local web file %s for writing!", 
				recvPath);
			return -1;
		}
		fclose(fd);
	}

	if (urlMode & URL_MODE_WRITE) {
		/* We must have at least one URL specified */
		if (myData == NULL || myData->urls == NULL) {
			uDebug(LOG_ERR, "At least one output URL must be specified!");
			return -1;
		}
	}

	/* if no interfaces defined, allocate dummy interface 
	 * out which we do broadcast.
 	 */
	if (!interfaces) {
		/* if we are using the bpf, we must have at least one
		 * interface.
		 */
#ifdef USE_BPF
            uDebug(LOG_ERR, "For BPF-based systems, you must put one interface in the config file");
			return -1;
#endif
        addIf("none", "b");
	    interfaces++;			
	}

	return 0;
}

int configReset()
{
	return 0;
}

int configDump()
{
	char typeName[25];

	if (senderType == NODE_TYPE_HA)
		strcpy(typeName, "Fixed Agent");
	else if (senderType == NODE_TYPE_MNODE)
		strcpy(typeName, "Mobile Node");
	else
		strcpy(typeName, "Unknown");
	uDebug(LOG_INFO, "The values loaded are:");
	uDebug(LOG_INFO, "Mode = %d", urlMode);
	if (urlMode & URL_MODE_WRITE) {
		uDebug(LOG_INFO, "Sender = %s ('%s')", senderName, typeName);
		uDebug(LOG_INFO, "Sender timeout = %d secs", sendTime);
		uDebug(LOG_INFO, "Sender Port = %d", sendPort);
		dumpURLs(myData);
	}
	if (urlMode & URL_MODE_READ) {
		uDebug(LOG_INFO, "Receiver Path = '%s'", recvPath);
		uDebug(LOG_INFO, "Expire = %d secs", expireTime);
		uDebug(LOG_INFO, "Receiver Port = %d", recvPort);
	}
	dumpIfs();

	return 0;
}

