/*
 *	URL Daemon : Project for Winter 2002
 *	Designed by 
 *		Sumit Chawla (sumit@cs.pdx.edu) 
 *	under guidance of
 *		Professor Jim Binkley (jrb@cs.pdx.edu)
 *
 *	File: urld.c
 *
 *	This is the main file, containing the loop that keeps the daemon
 *	running.
 *
 *	The first version of this software is quite trivial. We open a
 *	configuration file to read in the settings, and then start the
 *	daemon in one of the three modes: Read, Write or Interact. 
 *
 *	In Read mode, we keep listening for incoming messages on a 
 *	specific UDP port. Upon receipt of the message, we create an
 *	HTML file and write it to the specified path.
 *
 *	In Write mode, we keep sending our URL information periodically
 *	on the specified address and UDP port. It is expected that the
 *	address specified in the configuration file is a broadcast or
 *	multicast address.
 *
 *	In Interact mode, we do both of the above. We keep our own list 
 *	as well as that of other senders. BUT we send only our list, not
 *	of the others. This way, we can keep everything neat and clean.
 *
 *	We handle two kinds of signals. The first one takes in a request
 *	to stop; the main loop keeps checking this flag periodically. When this
 *	flag is set, the main loop stops the reader & writer, and exits 
 *	gracefully. The second is for USR1 signal, which forces a reload of the
 *	configuration file.
 *
**/

/* Headers */

#include <stdio.h>
#include <stdarg.h>
#include <signal.h>
#ifdef WIN32
#include <process.h>
#endif /* WIN32 */

#define URLD_MAIN_FILE
#include "urld.h"


/* Forward declarations */
int parseArgs(int argc, char **argv);
void usage();
void storePid();
extern int getReady();
extern int doWork();
#ifdef WIN32
extern int getopt(int argc, char *argv[], char *opts);
HANDLE hEventLog; 
#endif /* WIN32 */


/* 
	Here we go: Main function.
*/
int main(int argc, char **argv)
{
	/* Parse the arguments */
	if (parseArgs(argc, argv) < 0)
		return -1;

	/* Sanity check. We depend on this for getTL etc. */
	if (sizeof(unsigned short) != 2)
	{
		uDebug(LOG_ERR, "This platform doesn't have a 2-byte short, cannot proceed.");
		exit(-1);
	}

	/* Get ready for work */
	if (getReady() < 0) 
	{
		/* If there was an error, we'd have printed a message already. 
		*/
		return -1;
	}

#if !defined(WIN32) && !defined(sun)
	/* Become a daemon */
	if (!debug && !parseOnly) {
		if (getuid()) {
			fprintf(stderr, "Only superusers are allowed to run urld as daemon.\n");
			return -1;
		}

		if (daemon(0,0) < 0) 
		{
			fprintf(stderr,"Could not be 'daemon'ized, exiting\n");
			return -1;
		}
	}
#endif /* !defined(WIN32) && !defined(sun) */
	/* Store the PID */
	storePid();

	uDebug(LOG_INFO, "URL Web-server (%s) started as PID (%d)", URLD_VERSION, getpid());


	/* Start the infinite loop */
	return doWork();
}


int parseArgs (int argc, char **argv)
{
	extern int      errno;
	extern char    *optarg;
	extern int      optind;
	int             ch;

	strncpy(configFile, URLD_CONFIG_DEFAULT, 1024);
	while ((ch = getopt(argc, argv, "pwdc:")) != EOF) 
	{
		switch (ch) 
		{
		case 'p':
			parseOnly = 1;
		case 'd':
			debug = 1;
			break;
		case 'c':
			strncpy(configFile, optarg, 1024);
			break;
		case 'w':
			warnings = 1;
			break;
		case '?':
		default:
			usage();
			return -1;
		}
	}

	configFile[1023] = 0;
#ifdef WIN32
	if (!debug) {
		/* First, check if the registry entry for us exists or not. If not, create it */
		hEventLog = RegisterEventSource(NULL, "urld");
		if (hEventLog == NULL) {
			fprintf(stderr, "Could not register event source for URLD!");
			return -1;
		}
	}
#endif
	return 0;
}

void usage()
{
	fprintf(stderr,"Usage:\n\turld [-d] [-p] [-w] [-c <config file>]\n");
	fprintf(stderr, "Please see the urld man-page for details.\n");
}

#ifdef URLD_PIDFILE
void
storePid() {
	FILE *fp = fopen(URLD_PIDFILE, "w");

	/*
	 * No great loss if an error occurs -- the pid simply doesn't
	 * get stored.  Still should report it....   
	 */
	if (fp) {
		fprintf(fp, "%d\n", getpid());
		fclose(fp);
		return;
	}
	uDebug(LOG_ERR, "could not open %s", URLD_PIDFILE);
}
#else /* URLD_PIDFILE */
void storePid() {
	/* Do nothing */
}
#endif /* URLD_PIDFILE */


int uDebug(unsigned short f, char *format, ...)
{
	va_list ap;
	static char msgBuffer[1024];

	/* First step: log information messages only if debug flag is on */
	if (f == LOG_INFO && !debug) return 0;
	/* Second step: log warnings only if warning flag is enabled */
	if (f == LOG_WARN && !debug && !warnings) return 0;
	va_start(ap, format);
	/* vfprintf(fp, format, ap); */
	vsprintf(msgBuffer, format, ap);
	va_end(ap);
	if (debug) {
		fprintf(stderr, "%s\n", msgBuffer);
		fflush(stderr);
	} else {
#ifdef WIN32
		char *msgs[1];
		msgs[0] = msgBuffer;
		ReportEvent(hEventLog, f, 0, URLD_GENERAL_MESSAGE, NULL, 1, 0, msgs, NULL);
#else
		syslog(f, msgBuffer);
#endif
	}
	return 0;
}
