#ifndef URLD_HEADER
#define URLD_HEADER

/*
*   The stuff below declares the various directives. For Linux we don't
*   need anything specific. Win32 and Sun boxes use pthreads by default.
*   For BSD flavors, we need to define USE_BPF. I am printing a warning
*   if we don't detect a known configuration.
*/
#if defined(__FreeBSD__) || defined(FreeBSD) || defined(BSD)
#define USE_BPF
#elif defined(WIN32) 
#define THREADED_MODEL
#elif defined(sun)
#define THREADED_MODEL
#pragma message("Please ensure you are linking with pthreads library.")
#elif !defined(LINUX)
/* For safety, warn if we don't recognize this system */
#pragma message("If you need USE_BPF on this system, please modify the makefile")
#endif /* __FreeBSD__ */


#include <time.h>
#ifdef WIN32
#include <winsock.h>
#else /* WIN32 */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#endif /* WIN32 */
#ifdef THREADED_MODEL
#include <pthread.h>
#endif
#include "url_debug.h"

/*
 *	URL Daemon : Project for Winter 2002
 *	Designed by 
 *		Sumit Chawla (sumit@cs.pdx.edu) 
 *	under guidance of
 *		Professor Jim Binkley (jrb@cs.pdx.edu)
 *
 *	File: urld.h
 *
 *	This is the main header, containing all the common information.
 *
 *	For v1.0, I will concentrate only on BSD & Linux, leaving out the
 *	Windows-specific stuff for v1.1. But the structures should remain 
 *	unchanged across platforms.
 *
 *	Note that the structures mentioned below are how we store the data
 *	at runtime. For sending, the encoding used is pretty simple; we use
 *	TLV notation for sending the data.
**/

/* Most of the stuff below is pretty self-explanatory */
#define URLD_VERSION "2.0"

/* Note: it is not by accident that URL_MODE_INTERACT = UM_WRITE|UM_READ */
#define URL_MODE_DISABLE	0
#define URL_MODE_WRITE		1
#define URL_MODE_READ		2
#define URL_MODE_INTERACT	3

/* The various tags used for creating the TLV data */
#define URL_TAG_NODE_NAME	91
#define URL_TAG_URL			92
#define URL_TAG_DESCRIPTION 93
#define URL_TAG_NODE_TYPE	94
#define URL_TAG_END			99

/* The node type decides the format in HTML */
#define NODE_TYPE_UNK		0
#define NODE_TYPE_MNODE		1
#define NODE_TYPE_HA		2

/* Files by default on UNIX boxes; for Win, we will seek them in curdir */
#ifndef WIN32
#define URLD_CONFIG_DEFAULT "/etc/urld.conf"
#define URLD_PIDFILE		"/var/run/urld.pid"
#else /* WIN32 */
#define URLD_CONFIG_DEFAULT "urld.conf"
#define URLD_PIDFILE		"urld.pid"
#endif /* WIN32 */

/* Some constants that govern what the URLs etc can be */
#define URL_SIZE_MAX 128
#define URL_DESC_MAX 256
#define NODE_NAME_MAX 128
#define URL_MAXMSG 1500


/* The structure that contains URLs, both personal and remote */
typedef struct url_data {
	/* The URL itself, including the http:// stuff */
	char url[URL_SIZE_MAX];
	/* The English description of the URL. May include whitespace */
	char desc[URL_DESC_MAX];
	/* When was this URL last seen? */
	time_t tmval;
	/* Pointer to next URL */
	struct url_data *next;
} URL_DATA, *PURL_DATA;


/* The list of URLs and their senders */
typedef struct node_data {
	/* The sender */
	char name[NODE_NAME_MAX]; 
	int nodeType;
	char ipAddr[NODE_NAME_MAX];
	struct sockaddr_in socks;
	/* List of URLs */
	PURL_DATA urls;
	/* 
		When did we last hear from this guy? Or, for the sender,
		when did we send the last broadcast?
	*/
	time_t tmval;
	/* Next sender/receiver */
	struct node_data *next;
} NODE_DATA, *PNODE_DATA;


/* Global Data */
#ifdef URLD_MAIN_FILE
/* Global data read from the configuration file */
int senderType;
char senderName[NODE_NAME_MAX];
char recvPath[1024];
/* By default, use 30 seconds for timeout */
int expireTime = 30;
/* Ditto for "Refresh" pragma command */
int urlTime = 30;
/* use 1/3 of expire timeout for send time */
int sendTime = 10;
/* By default, we use UDP port 43333 */
unsigned short sendPort = 43333;
unsigned short recvPort = 43333;

/* By default we do broadcast */
int defaultMcast = 0;
/* Maximum number of URLs per system */
int maxURLs = 50;
/* Maximum number of systems per page */
int maxSystems = 10;

/* Colors we use in the HTML file */
const char *URLColors[] = {
	"0xff0000",
	"0x00ff00",
	"0x0000ff"
};


/* Stuff read from the command-line */
char configFile[1024];
int debug = 0;
int parseOnly = 0;
int warnings = 0;
int urlMode = URL_MODE_INTERACT;
PNODE_DATA myData = NULL; /* This is the stuff we read from config. */
PNODE_DATA remoteData = NULL; /* This is the stuff we get from others. */
int requestToStop = 0; /* If set, the main loop exits */

#else /* URLD_MAIN_FILE */
extern int senderType;
extern char senderName[];
extern char recvPath[];
/* By default, use 30 seconds for timeout & beacon */
extern int expireTime;
extern int urlTime;
extern int sendTime;
/* By default, we use UDP port 43333 */
extern unsigned short sendPort;
extern unsigned short recvPort;
extern const char *URLColors[];
extern int defaultMcast;
extern int maxURLs;
extern int maxSystems;

extern char configFile[];
extern int debug;
extern int parseOnly;
extern int warnings;
extern int urlMode;
extern PNODE_DATA myData;
extern PNODE_DATA remoteData;
extern int requestToStop;
#endif /* URLD_MAIN_FILE */

#ifdef WIN32
#define sleep(n) Sleep(n*1000)
#endif /* WIN32 */

#endif /* URLD_HEADER */
