/*
 *	URL Daemon : Project for Winter 2002
 *	Designed by 
 *		Sumit Chawla (sumit@cs.pdx.edu) 
 *	under guidance of
 *		Professor Jim Binkley (jrb@cs.pdx.edu)
 *
 *	File: urlstate.c
 *
 *	All the control functions are in this file, while not tying to a single
 *	implementation of urlcomm.
 *
**/
#include <stdio.h>
#include "urld.h"
#ifndef WIN32
#include <unistd.h>
#include <errno.h>
#endif /* WIN32 */
#include <signal.h>
/* CHANGED: We include our own MD5 files in URLD */
#include "md5.h"

/* The last hash */
static char lastHash[16];

/* Forward declarations */
int configReset();
int configDump();
int configRead(char *fileName);
int startupReader();
int startupWriter();
int stopReader();
int stopWriter();
int init_interfaces();
int anyChanges();
int writeToHTML();

#ifndef THREADED_MODEL
#include <signal.h>
int expireCounter = 0;
int sendCounter = 0;

void mnTick(int t)
{
	if (expireCounter > 0) expireCounter--;
	if (sendCounter > 0) sendCounter--;
	alarm(1);
}
#endif /* THREADED_MODEL */

/* For all callers, we attempt to close down gracefully. */
void onhup(int t)
{
	char	*message = "Daemon exiting (urld) - received HUP signal.";

	/* Close sockets, kill threads etc. */
	stopReader();
	stopWriter();
	/* Shut down gracefully */
	uDebug(LOG_INFO, message);
	exit(0);
}  /* end onhup() */



/*
 *	This function is only for resetting information stored as configuration.
 *	We don't do anything for communication; that must be handled explicitly
 *	by doWork().
*/
int getReady()
{
	int rc;
#ifdef WIN32
    WSADATA       wsd;
	/* We need to do this early, so that default interface can be created in configRead */
    if (WSAStartup(MAKEWORD(1,1), &wsd) != 0)
    {
        uDebug(LOG_ERR, "Failed to load Winsock!");
        return 1;
    }
#endif /* WIN32 */
	/*
		Release the information stored earlier, if applicable.
	*/
	configReset();
	/* 
		Load the configuration data 
	*/
	rc = configRead(configFile);
	if (rc < 0) 
	{
		configReset();
		return rc;    /* NOTE: all errors logged within configRead() */
	}

	if (parseOnly) 
	{
		debug = 1;
	}

	if (debug) 
	{
		configDump();
	}

	if (parseOnly) 
	{
		uDebug(LOG_INFO, "Parse mode -- successfully completed, exiting");
		configReset();
		return -1;
	}

	/* For signal-based implementation, set up the handler here */
#ifndef THREADED_MODEL
	/* For pure writers, no signal handling is needed */
	if (urlMode != URL_MODE_WRITE)
	{
		sendCounter = sendTime;
		expireCounter = expireTime;
		if (SIG_ERR == signal(SIGALRM, mnTick))
		{
			uDebug(LOG_ERR, "Signal handler could not be installed");
			exit (-1);
		}
		/* We start the alarm after setting up rest of the stuff */
	}
#endif /* THREADED_MODEL */

#ifndef WIN32
	/* We use SIGHUP etc. to exit gracefully. */
	signal(SIGHUP, onhup);
#else
	signal(SIGABRT, onhup);
#endif
	signal(SIGINT, onhup);	/* allows cleanup even when debugging */
    signal(SIGTERM, onhup);

	return init_interfaces();
}

/*
	The "main" routine. We use various functions from urlcomm, so that we can 
	change urlcomm completely without getting affected.
*/
int doWork()
{
#ifndef THREADED_MODEL
	fd_set set;
#endif /* THREADED_MODEL */
	/*
		Start up the communication channels
	*/

	/* If interfaces are defined, we need to initialize them */
	if (urlMode & URL_MODE_READ) 
	{
		if (startupReader() < 0)
			return -1;
	}
	if (urlMode & URL_MODE_WRITE) 
	{
		if (startupWriter() < 0)
		{
			stopReader();
			return -1;
		}
	}

	/* Start the first alarm */
#ifndef THREADED_MODEL
	if (urlMode != URL_MODE_WRITE)
		alarm(1);
#endif

	/*
		Now, we wait for something to happen
	*/
#ifdef THREADED_MODEL
	while (1)
	{
#ifdef WIN32
		Sleep(INFINITE);
#else /* WIN32 */
		pause();
#endif /* WIN32 */
	}
#else /* THREADED_MODEL */
	/* Set up the wait */
	/* Much easier to do it this way */
	if (urlMode == URL_MODE_WRITE)
	{
		uDebug(LOG_INFO, "Starting Write loop");
		while (1)
		{
			/* Nothing to do but write and sleep */
			writeOnce();
			sleep(sendTime);
		}
	}
	else
	while (1)
	{

		/* Did the timer go off? */
		if (urlMode & URL_MODE_WRITE)
			if (sendCounter == 0)
			{
				writeOnce();
				sendCounter = sendTime;
			}
		if (urlMode & URL_MODE_READ)
			if (expireCounter == 0)
			{
				writeToHTML();
				expireCounter = expireTime;
			}
		if (dataAvailable(&set) < 0) 
		{
			if (errno == EINTR)
				continue;
			uDebug(LOG_ERR, "Fatal error in main loop : Error %d!", errno);
			/* No need to exit here. Let the code continue. */
			/* exit (-1); */
		}
		else
			readOnce();
	}
#endif /* THREADED_MODEL */
	return 0;
}



int addURL(const char *url, const char *desc)
{
	/* Create the new node and add the passed info to it */
	PURL_DATA newUrl = (PURL_DATA)malloc(sizeof(URL_DATA));
	PNODE_DATA newNode = NULL;
	if (newUrl == NULL) {
		uDebug(LOG_ERR, "Out of memory!!!");
		exit(-1);
	}
	strcpy(newUrl->url, url);
	strcpy(newUrl->desc, desc);
	newUrl->next = NULL;
	/* We are adding the URL to our own list */
	if (myData == NULL) {
		newNode = (PNODE_DATA)malloc(sizeof(NODE_DATA));
	    if (newNode == NULL) {
		    uDebug(LOG_ERR, "Out of memory!!!");
		    exit(-1);
	    }
		newNode->next = NULL;
		newNode->urls = newUrl;
		myData = newNode;
	} else {
		/* myData already exists, insert this node */
		PURL_DATA eUrl = myData->urls;
		while (eUrl->next) eUrl = eUrl->next;
		eUrl->next = newUrl;
	}

	return 0;
}

int dumpURLs(PNODE_DATA nodeHead)
{
	PURL_DATA url;

	if (nodeHead == NULL) return 0;
	url = nodeHead->urls;
	while (url) {
		uDebug(LOG_INFO, "%s ('%s')", url->url, url->desc);
		url = url->next;
	}
	return dumpURLs(nodeHead->next);
}



char *getTL(unsigned short tag, unsigned short length)
{
	static char localBuffer[10];
	char *ptr = localBuffer;

	tag = htons(tag);
	length = htons(length);

	memcpy(ptr, &tag, 2);
	ptr += 2;
	memcpy(ptr, &length, 2);

	return localBuffer;
}

void getData(char *tl, unsigned short *tag, unsigned short *len)
{
	unsigned short t, l;
	memcpy(&t, tl, 2);
	memcpy(&l, tl + 2, 2);

	*tag = ntohs(t);
	*len = ntohs(l);
}

/*
*	Convert received Tag-Length-Value format to data node, merging
*	with existing values. If data is NULL, we create a new one.
*/
int tlvToData(char *buffer, int totalLength, PNODE_DATA *data, const char *ipAddr)
{
	PNODE_DATA curNode, lastNode;
	PURL_DATA curUrl, lastUrl;
	char *ptr = buffer;
	char *tempBuffer;
	int largestLen = 0;
	int urlCount = 0;
	int activeSystems = 0;
	
	unsigned short tag, len;
	int traversed = 0; /* To keep track of where we are */
	unsigned short lastSeen = 0; /* What tag did we just see? */
	int endOfUrls = 0;
    int newNodeFound = 0; /* If we added a new node. */

	/*
		First pass is sanity-check. We must have data in TLV format, 
		as follows:
			(N(UD)*)+E
			i.e a Node, followed by zero or more URL & Description pairs, 
		followed by more Nodes and their info, followed by the END tag.
	*/
	getData(ptr, &tag, &len);
	if (tag != URL_TAG_NODE_NAME) {
		uDebug(LOG_WARN, "First tag must be Node name! It is %d instead", tag);
		return -1;
	}
	if (len == 0) {
		uDebug(LOG_WARN, "Node name must be non-zero in length, not %d.", len);
		return -1;
	}
	traversed = 4 + len;
	ptr += traversed;
	largestLen = len;
	while (traversed < totalLength)
	{
		endOfUrls = 0;
		lastSeen = URL_TAG_NODE_NAME;
		while (!endOfUrls && traversed < totalLength)
		{
			getData(ptr, &tag, &len);
			/* Security check */
			if ((len == 0 && tag != URL_TAG_END) || (traversed + 4 + len > totalLength)) {
				/* Length is wrong, or corrupted. */
				uDebug(LOG_WARN, "Illegal length %d for tag %d.", tag, len);
				return -1;
			}
			switch(tag)
			{
			case URL_TAG_NODE_NAME:
				if (lastSeen != URL_TAG_NODE_NAME)
				{
					uDebug(LOG_WARN, "A Node name tag must follow a Node tag, not %d", lastSeen);
					return -1;
				}
				lastSeen = URL_TAG_NODE_NAME;
				traversed += (4 + len);
				ptr += (4 + len);
				if (len > largestLen) largestLen = len;
				endOfUrls++;
				break;
			case URL_TAG_NODE_TYPE:
				if (lastSeen != URL_TAG_NODE_NAME)
				{
					uDebug(LOG_WARN, "A Node type tag must follow a Node name tag, not %s", lastSeen);
					return -1;
				}
				if (len != 1)
				{
					uDebug(LOG_WARN, "Node type must be exactly one byte");
					return -1;
				}
				lastSeen = URL_TAG_NODE_TYPE;
				traversed += (4 + len);
				ptr += (4 + len);
				if (len > largestLen) largestLen = len;
				break;
			case URL_TAG_URL:
				if (lastSeen != URL_TAG_NODE_TYPE && lastSeen != URL_TAG_DESCRIPTION)
				{
					uDebug(LOG_WARN, "A URL tag must follow a Type/description tag, not %d", lastSeen);
					return -1;
				}
				lastSeen = URL_TAG_URL;
				traversed += (4 + len);
				ptr += (4 + len);
				if (len > largestLen) largestLen = len;
				urlCount++;
				break;
			case URL_TAG_DESCRIPTION:
				if (lastSeen != URL_TAG_URL)
				{
					uDebug(LOG_WARN, "A description tag must follow a URL tag, not %d", lastSeen);
					return -1;
				}
				lastSeen = URL_TAG_DESCRIPTION;
				traversed += (4 + len);
				ptr += (4 + len);
				if (len > largestLen) largestLen = len;
				break;
			case URL_TAG_END:
				if (lastSeen != URL_TAG_NODE_TYPE && lastSeen != URL_TAG_DESCRIPTION)
				{
					uDebug(LOG_WARN, "An end tag must follow a Type/Description tag, not %d", lastSeen);
					return -1;
				}
				if (len != 0)
					uDebug(LOG_WARN, "The end tag length should be zero..");
				lastSeen = URL_TAG_END;
				traversed += 4;
				/* Use this as the actual length we are interested in */
				totalLength = traversed;
				endOfUrls++;
				break;
			default:
				uDebug(LOG_WARN, "Unknown tag %d, skipping.", tag);
				/* This is OK, we should ignore this one. */
				lastSeen = tag;
				traversed += (4 + len);
				ptr += (4 + len);
				break;
			}
		}
	}

	/* The last tag must be URL_TAG_END */
	if (lastSeen != URL_TAG_END) {
		uDebug(LOG_WARN, "End tag missing, possibly corrupt message.");
		return -1;
	}

	/* Too many URLs? Not good.. */
	if (urlCount > maxURLs) {
		uDebug(LOG_WARN, "This message has %d URLs, while URLD is configured to handle only %d URLs per message",
			urlCount, maxURLs);
		return -1;
	}
	

	/*
		Second pass is for each of the nodes discovered. We search for the
		node name, and then for URLs within the node name. If a match occurs,
		we update the timestamp. If the node or the URL does not exist, we 
		create the new Node/URL.
		Since we already checked the form of the message, we can now simplify
		the traversal by trusting the TLV structure.
	*/
	ptr = buffer;
	traversed = 0;
	endOfUrls = 0;
	curNode = lastNode = NULL;
	curUrl = lastUrl = NULL;
	/* Allocate a buffer large enough to hold the longest string */
	tempBuffer = (char *)malloc(largestLen + 1);
    if (tempBuffer == NULL) {
        uDebug(LOG_ERR, "Out of memory while trying to allocate %d bytes", largestLen);
        exit(-1);
    }
	while (traversed < totalLength && !endOfUrls) 
	{
		getData(ptr, &tag, &len);
		ptr += 4;
		/* We do not care about unknown tags, but we still need to skip them */
		if (len && (len <= largestLen)) {
			strncpy(tempBuffer, ptr, len);
			tempBuffer[len] = 0;
		}
		switch (tag)
		{
			case URL_TAG_NODE_NAME:
				/* Start searching from beginning */
				curNode = *data;
				lastNode = NULL;
				activeSystems = 0;
				while (curNode)
				{
					activeSystems++;
					if (!strcmp(curNode->name, tempBuffer) && !strcmp(curNode->ipAddr, ipAddr))
						break;
					lastNode = curNode;
					curNode = curNode->next;
				}
				if (curNode == NULL)
				{
					/* If there are too many active systems, ignore messages */
					if (activeSystems >= maxSystems) {
						uDebug(LOG_ERR, "Ignoring message, too many systems in the vicinity.");
						free(tempBuffer);
						return -1;
					}
					/* Couldn't find the node, create a new one */
					curNode = (PNODE_DATA)malloc(sizeof(NODE_DATA));
                    if (curNode == NULL)   {
                        uDebug(LOG_ERR, "Out of memory!");
                        exit(-1);
                    }

					strcpy(curNode->name, tempBuffer);
					strcpy(curNode->ipAddr, ipAddr);
					curNode->next = NULL;
					curNode->urls = NULL;
					if (lastNode)
						lastNode->next = curNode;
					else
					{
						/* No nodes in the list, create a new one */
						*data = curNode;
					}
                    newNodeFound = 1;
				}
				/* Update the last time we saw this node */
				time(&(curNode->tmval));
				break;
			case URL_TAG_NODE_TYPE:
				if (tempBuffer[0] == NODE_TYPE_MNODE || tempBuffer[0] == NODE_TYPE_HA)
					curNode->nodeType = tempBuffer[0];
				else
					curNode->nodeType = NODE_TYPE_UNK;
				break;
			case URL_TAG_URL:
				curUrl = curNode->urls;
				lastUrl = NULL;
				while (curUrl)
				{
					if (strcmp(curUrl->url, tempBuffer) == 0)
						break;
					lastUrl = curUrl;
					curUrl = curUrl->next;
				}
				if (curUrl == NULL)
				{
					/* Couldn't find the node, create a new one */
					curUrl = (PURL_DATA)malloc(sizeof(URL_DATA));
                    if (curUrl == NULL) {
                        uDebug(LOG_ERR, "Out of memory!");
                        exit(-1);
                    }

					strcpy(curUrl->url, tempBuffer);
					curUrl->next = NULL;
					curUrl->desc[0] = 0;
					if (lastUrl == NULL)
						curNode->urls = curUrl;
					else
						lastUrl->next = curUrl;
                    newNodeFound = 1;
				}
				/* Either way, just update the timestamp */
				time(&(curUrl->tmval));
				break;
			case URL_TAG_DESCRIPTION:
				if (strlen(tempBuffer))
					strcpy(curUrl->desc, tempBuffer);
				else
					curUrl->desc[0] = 0;
				break;
			case URL_TAG_END:
				endOfUrls++;
				break;
			default:
				/* Do nothing, just skip this one */
				break;
		}
		traversed += (4 + len);
		ptr += len;
	}

	free(tempBuffer);
    if (newNodeFound) writeToHTML();
	return 0;
}

/*
*	Convert data to Tag-Length-Value format. Returns the length of TLV.
*/
int dataToTlv(PNODE_DATA data, char **tlv)
{
	PNODE_DATA pd = data;
	int totalLength = 0;

	/* Two-pass algorithm. First pass, we get the total length. Second pass, we form the TLV */
	int pass = 0;
	char *msg = NULL;
	char *tmp = NULL;
	unsigned short s;
	PURL_DATA pu = NULL;

	/* This stuff should work fine for multiple nodes in the single structure */
	for (pass = 0; pass < 2; pass++)
	{
		pd = data;
		if (pass == 1) tmp = msg;
		while (pd) 
		{
			switch (pass)
			{
			case 0:
				/* Estimate length of node info */
				totalLength += (4 + strlen(pd->name));
				/* Node type: 1 byte */
				totalLength += 5;
				pu = pd->urls;
				while (pu)
				{
					totalLength += (4 + strlen(pu->url));
					totalLength += (4 + strlen(pu->desc));
					pu = pu->next;
				}
				break;
			case 1:
				/* Copy stuff into the data */
				s = strlen(pd->name);
				memcpy(tmp, getTL(URL_TAG_NODE_NAME, s), 4);
				tmp += 4;
				strcpy(tmp, pd->name);
				tmp += s;
				memcpy(tmp, getTL(URL_TAG_NODE_TYPE, 1), 4);
				tmp += 4;
				tmp[0] = pd->nodeType;
				tmp++;
				pu = pd->urls;
				while (pu)
				{
					s = strlen(pu->url);
					memcpy(tmp, getTL(URL_TAG_URL, s), 4);
					tmp += 4;
					strcpy(tmp, pu->url);
					tmp += s;
					s = strlen(pu->desc);
					memcpy(tmp, getTL(URL_TAG_DESCRIPTION, s), 4);
					tmp += 4;
					strcpy(tmp, pu->desc);
					tmp += s;
					pu = pu->next;
				}
				break;
			}
			pd = pd->next;
		}
		/* If this is the first pass, prepare for next. Otherwise, we are done. */
		switch (pass)
		{
		case 0:
			if (totalLength == 0) return -1;
			/* Add the length of the end node to the total */
			totalLength += 4;
			msg = (char *)malloc(totalLength);
            if (msg == NULL) {
                uDebug(LOG_ERR, "Out of memory while trying to allocate %d bytes", totalLength);
                exit(-1);
            }

			break;
		case 1:
			/* Insert the terminating tag */
			memcpy(tmp, getTL(URL_TAG_END, 0), 4);
			*tlv = msg;
			break;
		}
	}
	return totalLength;
}

void htmlIntro(FILE *fd)
{
	fprintf(fd,"<html>\n");
	fprintf(fd,"<head>\n"); 
	fprintf(fd,"<title> URL Daemon: Listening on port %d </title>\n", recvPort);
	fprintf(fd,"</head>\n");
	fprintf(fd,"<body bgcolor=\"#FFFFFF\">\n");
	fprintf(fd,"<META HTTP-EQUIV=\"Content-Type\" CONTENT=\"text/html\"; charset=iso-8859-1>\n");
	fprintf(fd,"<META HTTP-EQUIV=\"Refresh\" CONTENT=\"%d\" >\n", urlTime);
	fprintf(fd,"<META HTTP-EQUIV=\"Pragma\" CONTENT=\"no-cache\" >\n");
	fprintf(fd,"<META HTTP-EQUIV=\"Expires\" CONTENT=\"Fri, 19 May 1999 23:37:00 GMT\">\n");
}

void htmlClose(FILE *fd)
{
	time_t ltime;

	time(&ltime);
	fprintf(fd, "<BR><BR><BR><BR><H4><CENTER>Last updated at : %s</CENTER></H4><BR>", ctime(&ltime));
	fprintf(fd,"</body></html>\n");
}

/*
    The writeToHTML function must be called with mutex acquired in thread-based
    implementation. It is currently called from either readOnce or from the 
    walker thread. Both of these have the mutex acquired.
*/
int writeToHTML()
{
	PNODE_DATA curNode;
	PURL_DATA curUrl;
	FILE *fd;
	int activeSystems = 0;

    if (!anyChanges()) return 0;

	fd = fopen(recvPath, "w");
	if (fd == NULL)
	{
		uDebug(LOG_ERR, "Uh-oh, can't open %s for writing", recvPath);
		return -1;
	}
	/* Write the header */
	htmlIntro(fd);
	curNode = remoteData;
	while (curNode)
	{
		activeSystems++;
		fprintf(fd, "<FONT COLOR=\"%s\">", URLColors[curNode->nodeType]);
		fprintf(fd, "<H2>%s (%s)</H2>\n<ul>\n", curNode->name, curNode->ipAddr);
		curUrl = curNode->urls;
		while (curUrl)
		{
			fprintf(fd, "<li><a href=\"%s\"> %s</a>\n", curUrl->url, curUrl->desc);
			curUrl = curUrl->next;
		}
		fprintf(fd, "</ul>\n");
		fprintf(fd, "</FONT>");
		curNode = curNode->next;
	}

	htmlClose(fd);
	fclose(fd);
	if (activeSystems > maxSystems) {
		uDebug(LOG_ERR, "Bug in code; we just wrote %d systems in one page, max was %d", 
					activeSystems, maxSystems);
	}
	return 0;
}


/*
	This function does three important things:
	1. Checks for expired tags, and marks them, so they are not written to HTML.
	2. Removes expired Urls/Nodes if they have exceeding NODE_TIMEOUT
	3. Creates the hash of Urls that do need to be written.

	It returns nonzero to indicate if step 3 returned a hash different from the
	one stored previously. A return value of zero indicates there's nothing to be
	written.
*/
int anyChanges()
{
	MD5_CTX context;
	static int firstTime = 1;
	PNODE_DATA curNode, prevNode;
	PURL_DATA curUrl, prevUrl;
	time_t rightNow;
	int retVal = 0;
	char newHash[16];

	/* If remoteData is NULL, we write a blank page, deleting previous one. */
	if (remoteData == NULL) 
	{
		firstTime = 1;
		return 1;
	}

	time(&rightNow);
	curNode = remoteData;
	prevNode = NULL;

	/* First, remove all nodes that we haven't heard from */
	while (curNode)
	{
		if (difftime(rightNow, curNode->tmval) >  expireTime)
		{
			/* Remove the URLs first */
			if (curNode->urls)
			{
				curUrl = curNode->urls;
				while (curUrl)
				{
					prevUrl = curUrl->next;
					free(curUrl);
					curUrl = prevUrl;
				}
			}
			/* Now remove the Node itself */
			if (prevNode)
				prevNode->next = curNode->next;
			else
				remoteData = curNode->next;
			prevNode = curNode->next;
			free(curNode);
			curNode = prevNode;
		}
		else
		{
			prevNode = curNode;
			curNode = curNode->next;
		}
	}

	/* Next, remove URLs individually as needed */
	curNode = remoteData;
	if (remoteData == NULL)
	{
		firstTime = 1;
		return 1;
	}
	while (curNode)
	{
		/* Same algo, this time for URLs instead of nodes */
		curUrl = curNode->urls;
		prevUrl = NULL;
		while (curUrl)
		{
			/* Has this URL expired already? */
			if (difftime(rightNow, curUrl->tmval) > expireTime)
			{
				/* Remove this URL */
				if (prevUrl)
					prevUrl->next = curUrl->next;
				else
					curNode->urls = curUrl->next;
				prevUrl = curUrl->next;
				free(curUrl);
				curUrl = prevUrl;
			}
			else
			{
				prevUrl = curUrl;
				curUrl = curUrl->next;
			}
		}
		/* Go to next node */
		curNode = curNode->next;
	}
	uDebug(LOG_INFO, "Finished pruning the node tree");

	/* Now walk through the stuff and calculate the new hash */
	curNode = remoteData;
	if (curNode == NULL)
	{
		firstTime = 1;
		return 1;
	}

	MD5_Init(&context);
	while (curNode)
	{
		MD5_Update(&context, curNode->name, strlen(curNode->name));
		curUrl = curNode->urls;
		while (curUrl)
		{
			MD5_Update(&context, curUrl->url, strlen(curUrl->url));
			MD5_Update(&context, curUrl->desc, strlen(curUrl->desc));
			curUrl = curUrl->next;
		}
		curNode = curNode->next;
	}
	

    MD5_Final(newHash, &context);
	/* Otherwise, check if the hash changed */
	if (firstTime)
	{
		firstTime = 0;
		uDebug(LOG_INFO, "Storing hash");
		memcpy(lastHash, newHash, 16);
		return 1;
	}
	else
	{
		int retVal = memcmp(lastHash, newHash, 16);
		memcpy(lastHash, newHash, 16);
		return retVal;
	}
}
