// We keep track of the node information in this class.
import java.util.*;
import java.io.*;

class UNode {
	Vector URLs;
	Vector Descs;
	Vector urlTS;
	String IPAddress;
	String Name;
	int nodeType;
	boolean doBroadcast;
	boolean doMulticast;
	UDebug debug;
	
	// For timestamps, use Calendar
	Calendar ts;
	
	public UNode(UDebug d)
	{
		URLs = new Vector();
		Descs = new Vector();
		urlTS = new Vector();
		IPAddress = "";
		Name = "";
		nodeType = 0; // Unknown
		doBroadcast = doMulticast = true;
		ts = Calendar.getInstance();
		debug = d;
	}
	
	public int getType() { return nodeType; }
	public synchronized void setType(int t) { nodeType = t;}
	
	public String getTypeColor() 
	{
		return  nodeType == UConf.NODE_TYPE_HA ? "0x00ff00" : 
										(nodeType == UConf.NODE_TYPE_MNODE ? "0x0000ff" : "0xff0000");
	}

	public String getIPAddress() { return IPAddress; }
	public synchronized void setIPAddress(String t) { IPAddress = t;}

	public String getName() { return Name; }
	public synchronized void setName(String t) { Name = t;}
	
	public Calendar getTS() { return ts; }
	public synchronized void setTS(Calendar ts) { this.ts = ts; }
	
	public int getURLCount() { return URLs.size(); }
	
	public synchronized void insert(String url, String desc)
	{
		URLs.addElement(url);
		Descs.addElement(desc);
		urlTS.addElement(Calendar.getInstance());
	}
	
	public String getURL(int i) { return (String)URLs.elementAt(i); }
	public String getDesc(int i) { return (String)Descs.elementAt(i); }
	public Calendar getTS(int i) { return (Calendar)urlTS.elementAt(i); }
	
	public synchronized void updateTS(int i, Calendar ts) { urlTS.setElementAt(ts, i); }
	
	public synchronized void delete(int i)
	{
		URLs.removeElementAt(i);
		Descs.removeElementAt(i);
		urlTS.removeElementAt(i);
	}
	
	public synchronized void deleteAll()
	{
		URLs.removeAllElements();
		Descs.removeAllElements();
		urlTS.removeAllElements();
		setName("");
		nodeType = 0;
	}
	
	private short getShort(ByteArrayInputStream b)
	{
		return (short)((b.read() << 8) + b.read());
	}
	
	// Based on the input, update the data of this node
	public boolean decode(byte[] buffer)
	{
		boolean endFound = false;
		
		// First and foremost, clear off all the stuff we have at the moment
		deleteAll();
		
		ByteArrayInputStream bais = new ByteArrayInputStream(buffer);
		short tag, length;
		byte[] data;
		String url = null, desc = null;
		
		// First TLV must always be the Node name
		tag = getShort(bais);
		if (tag != UConf.URL_TAG_NODE_NAME) {
			debug.print("Skipping this message, as the first tag is " + tag + ", not NODE_NAME");
			return false;
		}
		length = getShort(bais);
		if (length == 0){
			debug.print("Skipping this message, as the NODE_NAME length is 0");
			return false;
		}
		data = new byte[length];
		bais.read(data, 0, length);
		setName(new String(data));
		
		// Start the processing of the rest of the string
		int curPos = 0;
		while (bais.available() > 0) {
			tag = getShort(bais);
			length = getShort(bais);
			debug.print("Tag is " + Integer.toString(tag) + " and length is " + Integer.toString(length));
			switch (tag) {
			case UConf.URL_TAG_NODE_NAME:
				debug.print("Currently we allow only one name per message.");
				bais.skip(length);
				break;
			case UConf.URL_TAG_URL:
				data = new byte[length];
				bais.read(data, 0, length);
				url = new String(data);
				desc = null;
				break;
			case UConf.URL_TAG_DESCRIPTION:
				data = new byte[length];
				bais.read(data, 0, length);
				if (url == null) {
					debug.print("No URL specified for this desc ('"
								+ desc + "'), skipping");
					break;
				}
				desc = new String(data);
				insert(url, desc);
				url = desc = null;
				break;
			case UConf.URL_TAG_NODE_TYPE:
				data = new byte[length];
				bais.read(data, 0, length);
				setType(data[0]);
				break;
			case UConf.URL_TAG_END:
				debug.print("Found the end tag, skipping the rest of the bytes.");
				// Skip to the end of the stream
				bais.skip(bais.available());
				endFound = true;
				break;
			default:
				debug.print("Unknown tag " + tag + ", skipping..");
				// May be due to a bug in C code; if tag = length = 0, it's the end.
				if (length == 0) endFound = true;
				bais.skip(length);
				break;
			}
			if (endFound) break;
		}
		if (!endFound) {
			debug.print("Could not locate the end tag, message may have been truncated.");
			return false;
		}
		if (url != null || desc != null) {
			debug.print("Malformed message, an orphaned URL/Description found!");
			return false;
		}
		return true;
	}
}
