import java.io.*;
import java.net.*;
import java.util.*;
import java.security.*;

class UReader extends Thread {
	UConf ourConf;
	UDebug debug;
	
	UReader ourselves = null;
	
	Vector OurNodes;
	
	private boolean exitNow = false;
	private boolean proceed = false;
	private boolean changed = false;
	
	private byte[] lastHash = null;
	
	// Socket used for the receiving message
	MulticastSocket ms;
	
	// The HTML Writer thread; writes the HTML stuff out
	private class HTMLWriter extends Thread {
		private int wait;
		private UDebug debug;
		private UReader parent;
		boolean keepRunning, writeNow;
		
		HTMLWriter(int waitTime, UReader p, UDebug d)
		{
			wait = waitTime*1000;
			parent = p;
			debug = d;
			keepRunning = true;
			writeNow = false;
			OurNodes = new Vector();
			this.start();
		}
		
		public void run() {
			debug.print("Starting the HTML Writer.");
			try {
				while (keepRunning)	{
					while (!writeNow) 
						sleep(5000);
					sleep(wait);
					if (writeNow)
					{
						debug.print("Calling the write method");
						parent.WriteOut();
					}
				}
			}
			catch (InterruptedException e) {
				debug.print("The HTML Writer thread was interrupted.");
			}
		}
		
		public synchronized void doPause()
		{
			writeNow = false;
		}
		
		public synchronized void doResume()
		{
			writeNow = true;
		}
		
		public synchronized void doStop()
		{
			keepRunning = false;
			writeNow = false;
			parent = null;
		}
		
	}
	
	HTMLWriter ht;
	
	UReader(UConf c, UDebug d) 
	{
		ourConf = c;
		debug = d;
		ourselves = this;
		changed = true;
		ms = null;
		OurNodes = new Vector();
		ht = new HTMLWriter(ourConf.ExpireTime, this, debug);
	}
	
	// Function to check/insert new nodes
	private void CheckAndInsert(String sender, byte[] data, int length)
	{
		synchronized(OurNodes) {
			if (length == 0) return;
			// Translate the stuff to URLs and Descriptions etc.
			UNode tempNode = new UNode(debug);
			if (!tempNode.decode(data))	{
				debug.print("Decoding of sent data failed.");
				tempNode = null;
				return;
			}
			tempNode.IPAddress = sender;
			// Search for the sender
			int index = -1;
			for (int i = 0; i < OurNodes.size(); i++) {
				UNode n = (UNode) OurNodes.elementAt(i);
				if (n.getIPAddress().equals(sender)) {
					index = i;
					break;
				}
			}
			if (index == -1) {
				// This is a new entry
				OurNodes.addElement(tempNode);
			} else {
				// Refresh the entries, add new ones as needed
				UNode foundNode = (UNode)OurNodes.elementAt(index);
				boolean found;
				for (int i = 0; i < tempNode.getURLCount(); i++) {
					found = false;
					for (int j = 0; j < foundNode.getURLCount(); j++) {
						// If we located it, update the timestamp.
						if (tempNode.getURL(i).equals(foundNode.getURL(j))) {
							foundNode.updateTS(j, Calendar.getInstance());
							found = true;
							break;
						}
					}
					// New node wasn't there, add it now.
					if (!found) 
						foundNode.insert(tempNode.getURL(i), tempNode.getDesc(i));
				}
				// Also update the rest of the info
				foundNode.setType(tempNode.getType());
				foundNode.setName(tempNode.getName());
				foundNode.setTS(Calendar.getInstance());
				// Now we can get rid of the node
				tempNode = null;
			}
		}
	}
	
	// Remove all nodes and/or URLs that have not been refreshed
	// Reminder: we have to count DOWN, since removal will change the indices
	private void removeExpired(Calendar ts)
	{
		for (int i = OurNodes.size()-1; i >= 0; i--) {
			UNode u = (UNode) OurNodes.elementAt(i);
			if (u.getTS().before(ts)) {
				u.deleteAll();
				OurNodes.removeElementAt(i);
			} else {
				// Check for the URLs
				for (int j = u.getURLCount() -1; j >= 0; j--) {
					if (u.getTS(j).before(ts)) 
						u.delete(j);
				}
				if (u.getURLCount() == 0) {
					u.deleteAll();
					OurNodes.removeElementAt(i);
					u = null;
				}
			}
		}
	}
	
	// Function called by the HTML Writer.
	private void WriteOut()
	{
		synchronized(OurNodes) {
			boolean isDifferent = true;
			byte[] newHash;
			// Remove all stuff not updated "ExpireTime" seconds ago
			Calendar pastTime = Calendar.getInstance();
			pastTime.add(Calendar.SECOND, -1*ourConf.ExpireTime);
			removeExpired(pastTime);
			debug.print("Calculating hash to see if there was any change from last time");
			if (OurNodes.isEmpty())
				newHash = null;
			else {
				try {
					MessageDigest md = MessageDigest.getInstance("MD5");
					for (int i = 0; i < OurNodes.size(); i++) {
						UNode n = (UNode)OurNodes.elementAt(i);
						md.update(n.getIPAddress().getBytes());
						md.update(n.getName().getBytes());
						for (int j = 0; j < n.getURLCount(); j++)
						{
							md.update(n.getURL(j).getBytes());
							md.update(n.getDesc(j).getBytes());
						}
					}
					newHash = md.digest();
				} catch (NoSuchAlgorithmException err)	{
					debug.print("MD5 not available, writing every time..");
					newHash = null;
				}
			}
			if (lastHash == null || newHash == null)
				isDifferent = true;
			else if (lastHash.equals(newHash))
				isDifferent = false;
		
			lastHash = newHash;
			if (!isDifferent)
				return;

			debug.print("Writing the stuff to disk..");
			try {
				StringWriter s = new StringWriter();
				writeHeader(s);
				writeHTML(s);
				writeTail(s);
				debug.print(s.toString());
				FileOutputStream fos = new FileOutputStream(ourConf.HTMLFile);
				if (fos != null) {
					fos.write(s.toString().getBytes());
					fos.flush();
					fos.close();
					debug.print("Wrote to file successfully.");
				} else {
					debug.print("Couldn't write to file");
				}
				s.close();
			} catch (IOException e) {
				debug.print("Exception while writing to file: " + e.toString());
			}
			debug.print("Finished writing.");
		}
	}
	
	// Routines to create the HTML file
	private void writeHeader(StringWriter sw) throws IOException
	{
		sw.write("<html>\n");
		sw.write("<head>\n"); 
		sw.write("<title> URL Daemon: Listening on port " + ourConf.Port + " </title>\n");
		sw.write("</head>\n");
		sw.write("<body bgcolor=\"#FFFFFF\">\n");
		sw.write("<META HTTP-EQUIV=\"Content-Type\" CONTENT=\"text/html\"; charset=iso-8859-1>\n");
		sw.write("<META HTTP-EQUIV=\"Refresh\" CONTENT=\"" + ourConf.ExpireTime + "\" >\n");
		sw.write("<META HTTP-EQUIV=\"Pragma\" CONTENT=\"no-cache\" >\n");
		sw.write("<META HTTP-EQUIV=\"Expires\" CONTENT=\"Fri, 19 May 1999 23:37:00 GMT\">\n");
	}
	
	private void writeHTML(StringWriter sw) throws IOException
	{
		UNode curNode;
		for (int i = 0; i < OurNodes.size(); i++) {
			curNode = (UNode) OurNodes.elementAt(i);
			sw.write("<FONT COLOR=\"" + curNode.getTypeColor() + "\">");
			sw.write("<H2>");
			sw.write(curNode.getName());
			sw.write(" (");
			sw.write(curNode.getIPAddress());
			sw.write(")</H2>\n<ul>\n");
			for (int j = 0; j < curNode.getURLCount(); j++) {
				sw.write("<li><a href=\"");
				debug.print("Writing about " + curNode.getURL(j));
				sw.write(curNode.getURL(j));
				sw.write("\"> ");
				sw.write(curNode.getDesc(j));
				sw.write("</a>\n");
			}
			sw.write("</ul>\n</FONT>\n");
		}
	}
	
	public String getTimeStamp()
	{
		StringBuffer sb = new StringBuffer();
		String months[] = {"Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"};
		Calendar cal = Calendar.getInstance();
		sb.append(months[cal.get(Calendar.MONTH)]);
		if(cal.get(Calendar.DATE) <= 9)
			sb.append(" 0" + cal.get(Calendar.DATE));
		else
			sb.append(" " + cal.get(Calendar.DATE));
		sb.append(" " + cal.get(Calendar.YEAR) + ", ");
		if(cal.get(Calendar.HOUR) <= 9)
			sb.append(" 0" + cal.get(Calendar.HOUR));
		else
			sb.append(" " +cal.get(Calendar.HOUR));
		if(cal.get(Calendar.MINUTE) <= 9)
			sb.append(":0" + cal.get(Calendar.MINUTE));
		else
			sb.append(":" + cal.get(Calendar.MINUTE));
		return(sb.toString());
	}
	
	
	private void writeTail(StringWriter sw) throws IOException
	{
		sw.write("<BR><BR><BR><BR><H4><CENTER>Last updated at :" + getTimeStamp() + 
				 "</CENTER></H4><BR></body></html>\n");
	}
	
	// Prepare the reader and HTML threads
	private boolean loadConfiguration()
	{
		boolean retVal = true;
		
		String htmlFile = ourConf.HTMLFile;
		
		if (htmlFile == null)
		{
			debug.print("The HTML file must be specified.");
			return false;
		}
		
		File f = new File(htmlFile);
		
		if (f.canWrite() == false)
		{
			debug.print("Cannot write to " + htmlFile);
			return false;
		}
		f = null;
		
		try {
			// Also initialize the sockets we need to use
			InetAddress group = InetAddress.getByName("224.0.0.1");
			ms = new MulticastSocket(ourConf.Port);
			ms.joinGroup(group);
		} catch (Exception e) {
			debug.print("An exception occurred: " + e.toString());
			retVal = false;
		}
		
		return retVal;
	}
	

	public void run() 
	{
		debug.print("Starting the reader thread..");
		byte[] b = new byte[1500];
		while (!exitNow) {
			try {
				synchronized(ourselves) {
					// Are we allowed to proceed?
					while (!proceed)
						wait();
				}
				debug.print("Started.");
				if (changed) {
					// Re-load the configuration
					if (loadConfiguration() == false) {
						debug.print("Exiting due to configuration error.");
						ourselves = null;
						return;
					}
					changed = false;
					debug.print("Loaded the configuration file successfully.");
				}
				// Keep reading and waiting.
				debug.print("Waiting for incoming packets.");
				DatagramPacket p = new DatagramPacket(b, b.length);
				try {
					ms.receive(p);
					debug.print("Received a packet!");
				} catch (IOException err) {
					debug.print("An exception occurred during receive");
				}
				if (p.getLength() > 0)	{
					debug.print("Received an incoming message, processing it");
					CheckAndInsert(p.getAddress().getHostAddress(), p.getData(), p.getLength());
				}
			}
			catch (InterruptedException e) {
				debug.print("Reader thread was interrupted.");
			}
		}
		ourselves = null;
	}
	
	public synchronized void pauseThread()
	{
		ht.doPause();
		proceed = false;
		if (ourselves == null)
		{
			debug.print("The Writer is not running.");
			return;
		}
		// No need to interrupt an already interrupted thread.
		if (ourselves.interrupted()) return;
		ourselves.interrupt();
	}
	
	public synchronized void resumeThread()
	{
		ht.doResume();
		proceed = true;
		if (ourselves == null)
		{
			debug.print("The configuration errors do not allow us to proceed.");
			return;
		}
		ourselves.notify();
	}
	
	public synchronized void stopCompletely()
	{
		ht.doStop();
		exitNow = true;
		if (ms != null)
		{
			try {
				InetAddress group = InetAddress.getByName("224.0.0.1");
				ms.leaveGroup(group);
			} catch (Exception e) {
				debug.print("Exception during stop: " + e.toString());
			}
			ms.close();
		}
		ms = null;
		ourselves.interrupt();
	}
}
