import java.io.*;
import java.net.*;


class UWriter extends Thread {
	UConf ourConf;
	UDebug debug;
	
	UWriter ourselves = null;
	
	private boolean exitNow = false;
	private boolean proceed = false;
	private boolean changed = false;
	
	// Multicast and Broadcast
	MulticastSocket ms;
	DatagramSocket bs;
	DatagramPacket dpb, dpm;
	
	String confBeingUsed;
	
	
	UWriter(UConf c, UDebug d) 
	{
		ourConf = c;
		debug = d;
		ourselves = this;
		changed = true;
		ms = null;
		bs = null;
		dpb = dpm = null;
	}
	
	private boolean encode(ByteArrayOutputStream b, short tag, String data)
	{
		boolean retVal = true;
		int length = data.length();
		
		b.write(tag >> 8);
		b.write(tag & 0xFF);
		b.write(length >> 8);
		b.write(length & 0xFF);
		try {
			b.write(data.getBytes());
		} catch (IOException e) {
			retVal = false;
		}
		
		return retVal;
	}
	
	// Create the message, as per the protocol, to send.
	private boolean loadConfiguration()
	{
		boolean retVal = true;
		String ourName = ourConf.us.getName();
		int urlCount = ourConf.us.getURLCount();
		
		if (ourName == null)
		{
			debug.print("We must have a unique name.");
			return false;
		}
		if (urlCount <= 0) 
		{
			debug.print("At least one URL must be specified.");
			return false;
		}

		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		// Create the message as per the protocol definition
		if (!encode(baos, ourConf.URL_TAG_NODE_NAME, ourName)) return false;
		baos.write(0);
		baos.write(ourConf.URL_TAG_NODE_TYPE);
		baos.write(0);
		baos.write(1);
		baos.write(ourConf.us.getType());

		// Add the URLs and their descriptions
		for (int i = 0; i < urlCount; i++)
		{
			if (!encode(baos, ourConf.URL_TAG_URL, ourConf.us.getURL(i))) return false;
			if (!encode(baos, ourConf.URL_TAG_DESCRIPTION, ourConf.us.getDesc(i))) return false;
		}
		
		// Finally, the end of the string
		baos.write(0);
		baos.write(ourConf.URL_TAG_END);
		baos.write(0);
		baos.write(0);
		
		// The message encoding has been completed. We are ready to go.
		byte[] data = baos.toByteArray();
		
		try {
			// Also initialize the sockets we need to use
			if (ourConf.us.doBroadcast)
			{
				InetAddress bcast = InetAddress.getByName("255.255.255.255");
				bs = new DatagramSocket();
				dpb = new DatagramPacket(data, data.length, bcast, ourConf.Port);			
			}
		
			if (ourConf.us.doMulticast)
			{
				InetAddress group = InetAddress.getByName("224.0.0.1");
				ms = new MulticastSocket();
				dpm = new DatagramPacket(data, data.length, group, ourConf.Port);
				// setTTL has been deprecated..
				ms.setTTL((byte)1);
			}
		} catch (IOException e) {
			debug.print("An exception occurred: " + e.toString());
			retVal = false;
		}
		
		return retVal;
	}
	
	private void sendURLs()
	{
		try {
			if (ourConf.us.doBroadcast)
			{
				debug.print("Doing broadcast.");
				bs.send(dpb);
			}
		
			if (ourConf.us.doMulticast)
			{
				debug.print("Doing multicast.");
				ms.send(dpm);
			}
		} catch (IOException e) {
		}
	}
	
	public void run() 
	{
		debug.print("Starting the writer thread..");
		while (!exitNow) {
			try {
				synchronized(ourselves) {
					// Are we allowed to proceed?
					while (!proceed)
						wait();
				}
				if (changed) {
					// Re-load the configuration
					if (loadConfiguration() == false) {
						debug.print("Exiting due to configuration error.");
						ourselves = null;
						return;
					}
					changed = false;
				}
				// Send the stuff
				sendURLs();
				// And we are ready to sleep
				sleep(1000*(ourConf.SendTime));
			}
			catch (InterruptedException e) {
				debug.print("Writer thread was interrupted.");
			}
		}
		ourselves = null;
	}
	
	public synchronized void pauseThread()
	{
		proceed = false;
		if (ourselves == null)
		{
			debug.print("The Writer is not running.");
			return;
		}
		// No need to interrupt an already interrupted thread.
		if (ourselves.interrupted()) return;
		ourselves.interrupt();
	}
	
	public synchronized void resumeThread()
	{
		proceed = true;
		if (ourselves == null)
		{
			debug.print("The configuration errors do not allow us to proceed.");
			return;
		}
		ourselves.notify();
	}
	
	public synchronized void stopCompletely()
	{
		exitNow = true;
		if (ms != null)
			ms.close();
		if (bs != null)
			bs.close();
		ms = null;
		bs = null;
		dpb = dpm = null;
		ourselves.interrupt();
	}
}
